"""

    ScatteringBaseCorrector.py

    Copyright (c) 2017-2024, SAXS Team, KEK-PF

"""
import numpy                as np
import copy
import logging
from matplotlib.patches     import Polygon
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from molass_legacy._MOLASS.SerialSettings         import get_setting, get_xray_picking, XARY_BASE_LINEAR, XARY_BASE_QUADRATIC, XARY_BASE_SPLINED
from molass_legacy.SerialAnalyzer.ElutionBaseCurve       import ElutionBaseCurve
from molass_legacy.Baseline.ScatteringBaseline     import ScatteringBaseline
from ScatteringBasecurve    import ScatteringBasecurve
from ScatteringBasespline   import ScatteringBasespline
from molass_legacy.KekLib.Affine                 import Affine
from DebugCanvas            import DebugCanvas
from DevSettings            import get_dev_setting

DEBUG_PLOT              = False
VERIY_SMALL_ANGLE_LIMIT = 0.02
SMALL_ANGLE_LIMIT       = 0.1
USE_CONSTANT_PERCENTILE = True
ENABLE_ZERO_BASE        = True
HPM_RATIO               = 0.8
BETTER_INTEGRATIVE_CURVE = True

if USE_CONSTANT_PERCENTILE:
    from scipy.interpolate  import LSQUnivariateSpline

from BasePercentileOffset   import base_percentile_offset

def compute_baseline_using_LPM_impl(baseline_type, num_iterations, i, y, p_final=None, k=None, curve=None, logger=None, suppress_log=True, debug=False):
    """
    moved due to:
        ImportError: cannot import name 'LPM_PERCENTILE' from partially initialized module 'Baseline.Baseline' (most likely due to a circular import)
    """
    from molass_legacy.Baseline.Baseline import LPM_PERCENTILE, integrative_curve, better_integrative_curve

    baseline_total  = np.zeros(len(y))

    if baseline_type == 5:
        integral = True
        orig_y = y.copy()
    else:
        integral = False

    if not suppress_log:
        logger.info("computing the baseline with baseline_type=%d", baseline_type)

    if baseline_type > XARY_BASE_SPLINED:
        # temporary fix when called in LPM+MF
        baseline_type = XARY_BASE_LINEAR

    if debug:
        if k is not None and k < 10:
            print( 'LPM_impl', [k] )

    if p_final is None:
        p_final = LPM_PERCENTILE

    if baseline_type < 4:
        baseline_type_exec = baseline_type
    else:
        baseline_type_exec = XARY_BASE_LINEAR

    for k in range( num_iterations ):
        if baseline_type_exec == XARY_BASE_LINEAR:
            # x requiring warning is intentionally suppressed here fot the time being
            sbl = ScatteringBaseline( y, curve=curve, logger=logger, suppress_warning=True )
            D, E  = sbl.solve( p_final=p_final )
            xray_baseline   = D*i + E
        elif baseline_type_exec == XARY_BASE_QUADRATIC:
            sbl = ScatteringBasecurve( y )
            F, D, E = sbl.solve( p_final=p_final )
            xray_baseline   = F*i**2 + D*i + E
        else:
            assert( baseline_type_exec == XARY_BASE_SPLINED )
            sbl = ScatteringBasespline( y )
            F, D, E = sbl.solve( p_final=p_final )
            xray_baseline   = sbl.get_baseline( i )

        y -= xray_baseline
        baseline_total += xray_baseline

    if integral:
        if BETTER_INTEGRATIVE_CURVE:
            end_slices = curve.get_end_slices()
            temp_baseline, convex = better_integrative_curve(orig_y, baseline_total, p_final, end_slices=end_slices)
            if convex and logger is not None:
                logger.info("integrative curve has been made using convex method.")
        else:
            temp_baseline = integrative_curve(orig_y, baseline_total)
        diff_baseline = temp_baseline - baseline_total
        y -= diff_baseline
        baseline_total = temp_baseline

    if False:
        import molass_legacy.KekLib.DebugPlot as plt
        plt.plot( y )
        plt.plot( xray_baseline )
        plt.show()

    return baseline_total

class ScatteringBaseCorrector:
    def __init__( self, jvector, qvector, intensity_array,
                    curve=None,
                    index=None,
                    affine_info=None,
                    inty_curve_y=None,
                    baseline_opt=None,
                    baseline_type=None,
                    need_adjustment=None,
                    parent=None, with_demo=False,):

        self.logger     = logging.getLogger( __name__ )
        self.parent     = parent
        self.show_plot  = DEBUG_PLOT or with_demo
        self.qvector    = qvector
        self.jvector    = jvector
        self.intensity_array    = intensity_array
        self.curve      = curve
        self.x_vector   = affine_info[0]
        self.x_base     = affine_info[1]
        self.x_adjust   = affine_info[2]
        npp = int( len(self.x_vector)*HPM_RATIO )
        self.hpm_i      = sorted( np.argpartition( self.x_vector, npp )[npp:] )
        self.top_x      = np.average( self.hpm_i )
        self.hpm_y      = self.x_vector[self.hpm_i]
        self.top_y      = np.average( self.hpm_y )
        self.inty_curve_y   = inty_curve_y
        self.lpm_params = None

        if baseline_opt is None:
            baseline_opt    = get_setting( 'xray_baseline_opt' )

        if baseline_type is None:
            baseline_type = get_setting( 'xray_baseline_type' )

        self.logger.info("ScatteringBaseCorrector starts with baseline_type=%d", baseline_type)

        if need_adjustment is None:
            need_adjustment = get_setting( 'xray_baseline_adjust' ) == 1

        knots   = np.linspace( 0, len(jvector), len(jvector)//10 )
        self.iknots  = knots[1:-1]

        ecurve = ElutionBaseCurve( inty_curve_y )
        self.size_sigma = ecurve.compute_size_sigma()
        self.logger.info( 'size_sigma=%g' % ( self.size_sigma ) )

        self.baseline_opt       = baseline_opt
        self.baseline_type = baseline_type
        if baseline_type == 5:
            end_slices = get_setting('manual_end_slices')
        else:
            end_slices = None
        if end_slices is None:
            end_slices = curve.get_end_slices()
        self.end_slices = end_slices
        self.need_adjustment    = need_adjustment
        self.num_iterations     = get_setting( 'correction_iteration' )
        self.log_xray_lpm_params = get_dev_setting('log_xray_lpm_params')

    def correct_with_matrix_base( self, mapped_info, progress_cb=None, return_base=False ):
        from molass_legacy.Baseline.BaseScattering import basescattering_correct

        self.logger.info("correct_with_matrix_base")

        # use_mpi=True is not yet implemented properly
        base = basescattering_correct(mapped_info, self.intensity_array, use_mpi=False, end_slices=self.end_slices)
        if return_base:
            return base

    def correct_all_q_planes( self, progress_cb=None, return_base=False, debug_obj=None ):

        self.logger.info("correct_all_q_planes with optins: %s", str((self.baseline_opt, self.baseline_type)))

        if return_base:
            e_base_list = []

        if self.baseline_opt == 0:
            self.logger.info( 'no baseline correction will be performed.' )

        self.lpm_params = []
        for i in range( len(self.qvector) ):

            if return_base:
                base = self.correct_a_single_q_plane( i, return_baseline=True )
                e_base_list.append( base )
                if debug_obj is not None:
                    debug_obj.plot(i, base)
            else:
                try:
                    plot_closure = self.correct_a_single_q_plane( i )
                    if return_base:
                        e_base_list.append( plot_closure )
                except:
                    # 
                    plot_closure = None

                if plot_closure is not None:
                    figsize = ( 16, 8 ) if is_low_resolution() else ( 16, 8 )
                    dc = DebugCanvas( "Debug Z, Zs", plot_closure, figsize=figsize, parent=self.parent )
                    dc.show()
                    if not dc.continue_:
                        break

            # intensity_array[:,i,1] -= baseline
            if progress_cb is not None:
                progress_cb( i )

        self.logger.info( 'correct_all_q_planes done with p_finals %.3g-%.3g.', self.lpm_params[0][1], self.lpm_params[-1][1])
        if False:
            import molass_legacy.KekLib.DebugPlot as plt
            lpm_params = np.array(self.lpm_params)
            plt.push()
            fig, ax = plt.subplots()
            ax.set_title("Offset Percentile Values used with %.3g sigma width" % self.size_sigma, fontsize=16)
            ax.plot(self.qvector, lpm_params[:,0]*100, label='noise level')
            ax.plot(self.qvector, lpm_params[:,1], label='offset percentile')
            ax.legend()
            fig.tight_layout()
            plt.show()
            plt.pop()

        self.lpm_params = None
        if return_base:
            return np.array( e_base_list )

    def correct_a_single_q_plane( self, i, suppress_update=False, plot_always=False, return_baseline=False ):
        if len(self.intensity_array.shape) == 3:
            Z_  = self.intensity_array[:,i,1]
        else:
            Z_  = self.intensity_array[:,i]
        if suppress_update:
            Z   = copy.deepcopy(Z_)
        else:
            Z   = Z_
        Z_orig  = copy.deepcopy(Z)

        spline  = LSQUnivariateSpline( self.jvector, Z_orig, self.iknots )
        noisiness = self.compute_noisiness(spline, Z_orig)
        p_final =  base_percentile_offset( noisiness, size_sigma=self.size_sigma )
        if self.lpm_params is not None:
            self.lpm_params.append((noisiness, p_final))
        if self.log_xray_lpm_params:
            self.logger.info('[%d] noisiness=%g p_final=%g' % (i, noisiness, p_final) )

        baseline_total  = self.compute_baseline_and_corrrect( i, Z, p_final )
        plot_closure    = self.do_adjustment( i, Z, Z_orig, spline, baseline_total, plot_always, return_baseline, p_final )
        return plot_closure

    def compute_noisiness(self, spline, Z):
        pp  = np.percentile( Z, [95, 5] )
        height = pp[0] - pp[1]
        # height = pp[0]*0.5
        noisiness = np.std( Z - spline(self.jvector) ) / height
        return noisiness

    def compute_baseline_and_corrrect( self, i, Z, p_final, noisiness=None ):

        if self.baseline_opt == 0:
            baseline_total  = np.zeros(len(Z))
            return baseline_total

        q   = self.qvector[i]
        j   = self.jvector

        if q < VERIY_SMALL_ANGLE_LIMIT:
            iterate = self.num_iterations * 2
        elif q < SMALL_ANGLE_LIMIT:
            iterate = self.num_iterations
        else:
            iterate = 1

        if i < 10:
            print( 'correcting', [i] )

        try:
            baseline_total = compute_baseline_using_LPM_impl(self.baseline_type, iterate, j, Z, p_final=p_final, k=i, curve=self.curve)

        except Exception as exc:
            raise RuntimeError( 'failed to solve baseline at Q[%d]=%g; %s' % (i, q, str(exc)) )
            # note that intensity_array correnction is not complete in this cas

        return baseline_total

    def do_adjustment( self, i, Z, Z_orig, spline, baseline_total, plot_always, return_baseline, p_final ):
        show_plot   = self.show_plot

        if self.need_adjustment or show_plot:
            # construct Affine trasnformation
            x_vector        = self.x_vector
            x_base          = self.x_base
            x_adjust        = self.x_adjust
            x_total_adjust  = x_base + x_adjust
            x_end_i         = len(x_base)-1
            affine_src_x    = [ 0, self.top_x, x_end_i ]
            affine_src_y    = [ x_base[0], self.top_y, x_base[-1] ]
            src_points      = list( zip( affine_src_x, affine_src_y ) )

            affine_tgt_x    = affine_src_x
            tgt_hpm_y       = spline(self.hpm_i)
            tgt_top_y       = np.average( tgt_hpm_y )
            affine_tgt_y    = [ baseline_total[0], tgt_top_y, baseline_total[-1] ]
            tgt_points      = list( zip( affine_tgt_x, affine_tgt_y ) )

        if self.need_adjustment:
            affine          = Affine( src_points, tgt_points )

            # affine transform only the end points
            x_adjust_ends   = x_total_adjust[ [0, -1] ]
            tx, ty          = affine.transform( self.jvector[[0, -1]], x_adjust_ends )

            # compute the linear adjustment vector in the target plane
            affined_line    = np.linspace( ty[0], ty[-1], len(x_base) )
            total_line      = np.linspace( baseline_total[0], baseline_total[-1], len(x_base) )
            affined_adjust  = affined_line - total_line

            Z -= affined_adjust     # note that this is updating intensity_array[:,i,1]
        else:
            affined_adjust  = 0

        if return_baseline:
            return baseline_total + affined_adjust

        if plot_always or show_plot and i%10==0:

            # compute the adjusted baseline in the target plane
            affined_total_adjust    = baseline_total + affined_adjust
            src_adjust_points   = [ (0, x_base[0]), (0, x_total_adjust[0]), (x_end_i, x_total_adjust[x_end_i]), (x_end_i, x_base[x_end_i]) ]
            tgt_adjust_points   = [ (0, baseline_total[0]), (0, affined_total_adjust[0]), (x_end_i, affined_total_adjust[x_end_i]), (x_end_i, baseline_total[x_end_i]) ]

            def plot_closure(fig, axes=None, same_scaling=False, zero_base=False, j0=0):
                zero_base = ENABLE_ZERO_BASE and zero_base

                if axes is None:
                    ax1 = fig.add_subplot( 121 )
                    ax2 = fig.add_subplot( 122 )
                else:
                    ax1, ax2 = axes

                total_adjust_color      = 'red'
                adjust_polygon_coloer   = 'pink'

                intensity_picking = get_xray_picking()
                ax1.set_title( 'Elution around Q=%g (averaged)' % ( intensity_picking ) )
                ax2.set_title( 'Elution at Q[%d]=%.3g' % ( i, self.qvector[i] ) )

                x = j0 + np.arange(len(Z))
                affine_src_x_ = list(j0 + np.array(affine_src_x))
                affine_tgt_x_ = list(j0 + np.array(affine_tgt_x))

                if zero_base:
                    zero_base_shift = x_total_adjust
                    w = self.top_y / x_end_i
                    src_peak_shift  = zero_base_shift[0] * ( 1 - w ) + zero_base_shift[x_end_i] * w
                    affine_src_y_   = [ affine_src_y[0] - zero_base_shift[0], self.top_y - src_peak_shift, affine_src_y[2] - zero_base_shift[x_end_i] ]
                    src_points_     = list( zip( affine_src_x_, affine_src_y_ ) )
                    src_adjust_points_  = [ (j0, x_base[0] -  zero_base_shift[0]), (j0, 0), (j0+x_end_i, 0), (j0+x_end_i, x_base[x_end_i] - zero_base_shift[x_end_i]) ]
                else:
                    zero_base_shift = 0
                    affine_src_y_   = affine_src_y
                    src_points_     = list( zip( affine_src_x_, affine_src_y_ ) )
                    # src_adjust_points_ = src_adjust_points
                    src_adjust_points_  = [ (j0, x_base[0]), (j0, x_total_adjust[0]), (j0+x_end_i, x_total_adjust[x_end_i]), (j0+x_end_i, x_base[x_end_i]) ]

                ax1.plot( x_vector - zero_base_shift, color='orange' )

                if self.need_adjustment:
                    ax1.plot( x, x_base - zero_base_shift, color='pink' )
                    ax1.plot( x, x_total_adjust - zero_base_shift, color=total_adjust_color, alpha=0.5 )
                else:
                    ax1.plot( x, x_base - zero_base_shift, color='red' )

                affine_point_colors = [ 'yellow', 'red', 'cyan' ]

                for k in range( 3 ):
                    ax1.plot( affine_src_x_[k], affine_src_y_[k], 'o', color=affine_point_colors[k], markersize=10 )

                src_polygon = Polygon( src_points_, alpha=0.2 )
                ax1.add_patch(src_polygon)

                if self.need_adjustment:
                    src_adjust_polygon = Polygon( src_adjust_points_, alpha=0.3, fc=adjust_polygon_coloer )
                    ax1.add_patch(src_adjust_polygon)

                def draw_text_bpo(ax, bpo):
                    pass

                if zero_base:
                    zero_base_shift = affined_total_adjust
                    w = self.top_y / x_end_i
                    tgt_peak_shift  = zero_base_shift[0] * ( 1 - w ) + zero_base_shift[x_end_i] * w
                    affine_tgt_y_   = [ affine_tgt_y[0] - zero_base_shift[0], tgt_top_y - tgt_peak_shift, affine_tgt_y[2] - zero_base_shift[x_end_i] ]
                    tgt_points_     = list( zip( affine_tgt_x_, affine_tgt_y_ ) )
                    tgt_adjust_points_  = [ (j0, baseline_total[0] - zero_base_shift[0]), (j0, 0), (j0+x_end_i, 0), (j0+x_end_i, baseline_total[x_end_i] - zero_base_shift[x_end_i]) ]
                else:
                    zero_base_shift = 0
                    affine_tgt_y_   = affine_tgt_y
                    # tgt_points_     = tgt_points
                    tgt_points_ = list( zip( affine_tgt_x_, affine_tgt_y ) )
                    # tgt_adjust_points_  = tgt_adjust_points
                    tgt_adjust_points_ = [ (j0, baseline_total[0]), (j0, affined_total_adjust[0]), (j0+x_end_i, affined_total_adjust[x_end_i]), (j0+x_end_i, baseline_total[x_end_i]) ]

                ax2.plot( x, Z_orig - zero_base_shift, color='orange' )
                ax2.plot( x, spline( self.jvector ) - zero_base_shift, ':', color='green' )
                if self.need_adjustment:
                    ax2.plot( x, baseline_total - zero_base_shift, color='pink' )
                    ax2.plot( x, affined_total_adjust - zero_base_shift, color=total_adjust_color, alpha=0.5 )
                else:
                    ax2.plot( x, baseline_total - zero_base_shift, color='red' )

                for k in range( 3 ):
                    ax2.plot( affine_tgt_x_[k], affine_tgt_y_[k], 'o', color=affine_point_colors[k], markersize=10 )

                tgt_polygon = Polygon( tgt_points_, alpha=0.2 )
                ax2.add_patch(tgt_polygon)

                if self.need_adjustment:
                    tgt_adjust_polygon = Polygon( tgt_adjust_points_, alpha=0.3, fc=adjust_polygon_coloer )
                    ax2.add_patch(tgt_adjust_polygon)

                if True:
                    if zero_base:
                        src_hpm_y_  = self.hpm_y - x_total_adjust[self.hpm_i]
                        tgt_hpm_y_  = tgt_hpm_y - affined_total_adjust[self.hpm_i]
                    else:
                        src_hpm_y_  = self.hpm_y
                        tgt_hpm_y_  = tgt_hpm_y

                    ax1.plot( j0+np.array(self.hpm_i), src_hpm_y_, 'o', color='red', markersize=3, alpha=0.5 )
                    ax2.plot( j0+np.array(self.hpm_i), tgt_hpm_y_, 'o', color='red', markersize=3, alpha=0.5 )

                # ymin, ymax = 

                if same_scaling:
                    ymin1, ymax1 = ax1.get_ylim()
                    ymin2, ymax2 = ax2.get_ylim()
                    ymin_ = min( ymin1, ymin2 )
                    ymax_ = max( ymax1, ymax2 )
                    for ax in [ax1, ax2]:
                        ax.set_ylim( ymin_, ymax_ )

                fig.tight_layout()

            ret = plot_closure
        else:
            ret = None

        return ret



"""
    StdoutLogger.py

    adapted from
    https://stackoverflow.com/questions/616645/how-to-duplicate-sys-stdout-to-a-log-file

    Copyright (c) 2020, Masatsuyo Takahashi, KEK-PF
"""
import sys

class StdoutLogger(object):
    def __init__(self, logfile="stdout.log"):
        self.terminal = sys.stdout
        self.log = open(logfile, "a", encoding='utf-8')
        sys.stdout = self

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)  

    def flush(self):
        self.terminal.flush()
        self.log.flush()

    def __del__(self):
        self.log.close()
        sys.stdout = self.terminal



# -*- coding: utf-8 -*-
"""

    ファイル名：   CommandLineOptions.py

    処理内容：

        コマンドライン・オプションの取得

"""
from __future__ import division, print_function, unicode_literals
import sys

from optparse import OptionParser

p = OptionParser()

# required
p.add_option('-c', '--command', action='store_true',
                help="run in command mode; default is GUI mode")

p.add_option('-i', '--in-folder', action='store', dest='in_folder',
                help="specify input image folder; only required option in command mode")

# optinal
p.add_option('-o', '--out-folder', action='store', dest='out_folder',
                help="specify output image folder; default is IN_FOLDER/Synthesized")

p.add_option('-n', '--autonum-folders', action='store_true', dest='autonum_folders',
                help="create automatically numbered output folder if default output folder exists")

p.add_option('-j', '--adj-folder', action='store', dest='adj_folder',
                help="specify adjusted image folder; default is no adjusted image output")

p.add_option('-m', '--intermediate-results', action='store_true', dest='intermediate_results',
                help="write intermediate results such as *_1_sun.tif")

p.add_option('-p', '--pandastable', action='store_true', 
                help="use pandastable in GUI")

p.add_option('-v', '--version', action='store_true', 
                help="show the appication's version info")

opts, args = p.parse_args()



# coding: utf-8
"""
    SecTheory.Distinction.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def hard_sphere_curves():
    from bisect import bisect_right
    from Theory.SolidSphere import phi
    from Theory.Rg import compute_Rg
    from Error.ErrorModel import error_model

    qv = np.linspace(0.005, 0.5, 500)

    R50, R30 = [np.sqrt(5/3)*r for r in [50, 30]]

    y50 = phi(qv, R50)**2
    y30 = phi(qv, R30)**2
    ymix = (y50+y30)/2

    i = bisect_right(qv, 0.01)
    for y in [y50, y30, ymix]:
        rg = compute_Rg(qv[0:i], y[0:i])
        print("Rg=", rg)

    k = 1e4
    c = 0.05
    e50, e30, emix = [error_model(qv, y, k, c, y[0]) for y in [y50, y30, ymix]]

    with plt.Dp():
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))

        ax1.set_title("Individual Curves", fontsize=20)
        ax2.set_title("Mixture Curve", fontsize=20)
        ax3.set_title("Error Curves", fontsize=20)

        for ax in [ax1, ax2, ax3]:
            ax.set_yscale("log")

        ax1.plot(qv, y50, label="Sphere Rg=50")
        ax1.plot(qv, y30, label="Sphere Rg=30")
        ax2.plot(qv, ymix, label="1:1 Mixture")

        ax3.plot(qv, e50, label="Rg=50")
        ax3.plot(qv, e30, label="Rg=30")
        ax3.plot(qv, emix, label="1:1 Mixture")

        for ax in [ax1, ax2, ax3]:
            ax.legend(fontsize=16)

        fig.tight_layout()
        plt.show()

    for name, rec in [
            ("rg50.dat", (qv, y50, e50)),
            ("rg30.dat", (qv, y30, e30)),
            ("mix.dat", (qv, ymix, emix))]:
        np.savetxt(name, np.array(rec).T)

def sec_elution_curves():
    from .BasicModels import SimpleSec

    rp0 = 200
    sec = SimpleSec(rp0=rp0, nperm=200, tperm=1)

    # x = np.arange(300)
    x = np.linspace(0.001, 300, 300)

    curves = []
    for rg in [50, 41, 30]:
        y  = sec.get_elution(x, rg)
        curves.append(y)

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))

        fig.suptitle("SEC Theory Prediction with Pore Size=%s" % rp0, fontsize=24)
        ax1.set_title("Separate Elutions", fontsize=20)
        ax2.set_title("Poly vs. Mono", fontsize=20)

        for y, name in zip(curves, ["rg50", "rg41", "rg30"]):
            ax1.plot(x, y, label=name)

        y1 = (curves[0] + curves[2])/2
        y2 = curves[1]
        ax2.plot(x, y1, label="Polydisperse (rg50, rg30)")
        ax2.plot(x, y2, label="Monodisperse (rg41)")

        for ax in [ax1, ax2]:
            ax.legend(fontsize=14)
        fig.tight_layout()
        plt.show()



"""
    OutlineViewer.py

    Copyright (c) 2018-2023, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.widgets import Button, SpanSelector
from matplotlib.patches import Polygon
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import get_color, NavigationToolbar
from molass_legacy.KekLib.TkUtils import is_low_resolution
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

USE_PLOT_SURFACE = False

def format_coord(x, y):
    return 'x=%.4f    y=%.4f' % (x, y)

def using_normal_conc():
    return get_setting('use_xray_conc') == 0 and get_setting('use_mtd_conc') == 0

class OutlineViewer3dFrame(Tk.Frame):
    def __init__(self, parent, parent_obj, figsize, title, x, y, data, curve_info):
        self.in_folder = get_setting('in_folder')
        self.showing_xray = title.find('Xray') >= 0
        self.using_normal_conc = using_normal_conc()
        Tk.Frame.__init__(self, parent)
        cframe = Tk.Frame(self)
        cframe.pack()
        self.fig = plt.figure( figsize=figsize )
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        if self.showing_xray or self.using_normal_conc:
            self.ax = self.fig.add_subplot(111, projection='3d')

            if title is not None:
                self.ax.set_title(title, y=1.1)
            if data is None:
                x = np.linspace(0, 1, 10)
                self.ax.plot(x, x, x)
            else:
                self.draw_3d(x, y, data, curve_info)
        else:
            self.ax = self.fig.gca()
            self.ax.text(0.5, 0.5, "Not available", ha='center', fontsize=40, alpha=0.3)
            self.ax.set_axis_off()
        self.mpl_canvas.draw()

    def close(self):
        plt.close(self.fig)

    def draw_3d(self, x, y, data, curve_info):
        self.fig.suptitle(self.in_folder)
        color = get_color(0)

        if USE_PLOT_SURFACE:
            from MatrixData import simple_plot_3d
            simple_plot_3d(self.ax, data)
        else:
            skip_cycle = data.shape[0]//100
            for i in range(data.shape[0]):
                if i % skip_cycle != 0:
                    continue
                y_   = y
                x_   = np.ones(len(y))*x[i]
                z_   = data[i,:]
                self.ax.plot(x_, y_, z_, color=color, alpha=0.3)

        if curve_info is not None:
            for k, info in enumerate(curve_info):
                pos, values, c = info[0:3]
                linestyle = info[3] if len(info) > 3 else None
                force_eview = info[4] if len(info) > 4 else False   # too tricky, better refactor
                if k % 2 == 0 or force_eview:
                    x_ = np.ones(len(values))*x[pos]
                    y_ = y
                    z_ = values
                else:
                    x_ = x
                    y_ = np.ones(len(values))*y[pos]
                    z_ = values
                self.ax.plot(x_, y_, z_, color=c, linestyle=linestyle)

        pos = plt.axes([0.8, 0.05, 0.1, 0.04])
        self.reset_view_btn = Button(pos, 'Reset View')
        self.reset_view_btn.on_clicked(self.reset_view)

    def reset_view(self, event):
        self.ax.view_init()
        self.mpl_canvas.draw()

class OutlineViewer2dFrame(Tk.Frame):
    def __init__(self, parent, parent_obj, figsize, title, x, y, data, color, opos, ocolor, restrict_info,
                sub_curve_info=None, pre_recog=None, ecurve=None, opos_sub=None, disable_selection=False,
                changeable_yscale=False, bridge_slice=None, toolbar=False, val_label=False):
        self.title = title
        self.x = x
        self.y = y
        self.data = data
        self.color = color
        self.opos = opos
        self.ocolor = ocolor
        self.restrict_info = restrict_info
        self.sub_curve_info = sub_curve_info
        self.pre_recog = pre_recog
        self.ecurve = ecurve
        self.showing_uv = pre_recog is not None
        self.opos_sub = opos_sub
        self.disable_selection = disable_selection
        self.selection_guide_text = "Drag to suggest where to select"
        self.changeable_yscale = changeable_yscale
        self.bridge_slice = bridge_slice
        self.toolbar_flag = toolbar
        self.val_label_flag = val_label
        self.max_value = len(y) - 1
        Tk.Frame.__init__(self, parent)

        self.normal_plot = title.find('Xray') >= 0 or using_normal_conc() or title.find('Elution') >= 0
        self.cbvar = None

        cframe = Tk.Frame(self)
        cframe.pack()
        self.build_cframe(cframe, figsize)

        bframe = Tk.Frame(self)
        bframe.pack(fill=Tk.X, expand=1)

        if self.toolbar_flag:
            bframe_ = self.add_toolbar(bframe)
        else:
            bframe_ = bframe

        self.build_bframe(bframe_)

    def add_toolbar(self, bframe):
        tbframe = Tk.Frame(bframe)
        tbframe.pack(side=Tk.LEFT)
        self.ax.format_coord = format_coord     # override the default to avoid fluctuating
        # self.toolbar = NavigationToolbar( self.mpl_canvas, tbframe, show_mode=False )
        self.toolbar = NavigationToolbar( self.mpl_canvas, tbframe )
        self.toolbar.update()
        bframe_ = Tk.Frame(bframe)
        bframe_.pack(side=Tk.RIGHT)
        return bframe_

    def build_bframe(self, bframe_):
        pass

    def build_cframe(self, cframe, figsize):
        self.fig, self.ax = plt.subplots( figsize=figsize )
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        if self.normal_plot:
            if self.changeable_yscale:
                self.log_scale = 1
                self.button_ax = plt.axes([0.9, 0.9, 0.09, 0.08])
            else:
                self.log_scale = None
                self.button_ax = None

            if not self.disable_selection:
                self.span = SpanSelector(self.ax, self.onselect, 'horizontal', useblit=True,
                            props=dict(alpha=0.5))
            self.draw()
        else:
            self.ax.text(0.5, 0.5, "Not available", ha='center', fontsize=40, alpha=0.3)
            self.ax.set_axis_off()
            self.mpl_canvas.draw()

    def close(self):
        plt.close(self.fig)

    def draw(self, canvas_draw=True, restrict_info=None):
        ax = self.ax
        ax.cla()
        if self.changeable_yscale:
            self.button_ax.cla()
            if self.log_scale:
                scale = 'log'
                btn_text = 'To Linear'
            else:
                scale = 'linear'
                btn_text = 'To Log'
            ax.set_yscale(scale)
            self.toggle_scale_btn = Button(self.button_ax, btn_text)
            self.toggle_scale_btn.on_clicked(self.toggle_scale)
        else:
            pass

        ax.set_title(self.title)
        ax.plot(self.x, self.y, color=self.color, linewidth=3)
        if self.sub_curve_info is not None:
            i_sub, y_sub, c_sub, s_sub = self.sub_curve_info
            ax.plot(self.x, y_sub, color=c_sub, linestyle=s_sub, linewidth=1)

        ax.set_ylim(ax.get_ylim())

        if self.bridge_slice is None:
            ox = self.x[self.opos]
            oy = self.y[self.opos]
            ax.plot([ox, ox], [0, oy], color=self.ocolor, linewidth=3, alpha=0.5)
        else:
            start   = self.bridge_slice.start
            stop    = self.bridge_slice.stop + 1  # include stop
            slice_  = slice( start, stop )
            x_  = self.x[slice_]
            y_  = self.y[slice_]
            ymin, ymax = ax.get_ylim()
            verts = [ (x_[0], ymin) ] + list(zip(x_, y_)) + [(x_[-1], ymin)]
            poly = Polygon(verts, facecolor=self.ocolor, alpha=0.5 )
            ax.add_patch(poly)

        if self.opos_sub is not None:
            ox = self.x[self.opos_sub]
            oy = self.y[self.opos_sub]
            ax.plot([ox, ox], [0, oy], color=self.ocolor, linestyle=':', alpha=0.5)

        if not self.disable_selection:
            self.draw_selection(ax, restrict_info)

        if self.showing_uv:
            fc_list = self.pre_recog.flowchange.get_real_flow_changes()
            to_plot = []
            for fc in fc_list:
                if fc is not None:
                    to_plot.append(fc)
            ax.plot(to_plot, self.ecurve.spline(to_plot), 'o', color='cyan')
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            xoffset = (xmax - xmin)*0.1
            yoffset = (ymax - ymin)*0.15
            for k, fc in enumerate(fc_list):
                if fc is None:
                    continue

                fx = fc
                fy = self.ecurve.spline(fx)
                signed_scale = -1.5 if k == 0 else 0.5
                ax.annotate( "flow change", xy=( fx, fy ),
                    xytext=( fx + signed_scale*xoffset, fy + yoffset ), alpha=1,
                    arrowprops=dict( headwidth=5, width=1, color='black', shrink=0 ),
                    )

        self.fig.tight_layout()
        if canvas_draw:
            self.mpl_canvas.draw()

    def onselect(self, xmin, xmax):
        pass

    def draw_selection(ax, restrict_info):
        pass

    def toggle_scale(self, event):
        self.log_scale = 1 - self.log_scale
        self.toggle_scale_btn = None
        self.draw()

class OutlineViewerFrame(Tk.Frame):
    def __init__(self, parent, dialog, titles, x, y, data, epos, ecurve, el_color, restrict_list,
                    toggle_info, disable_x_selection=False,
                    averaged_ridge=False,
                    changeable_yscale=False, sub_curve_info=None, pre_recog=None,
                    x_axis_ends=None,
                    bridge_slice=None,
                    sub_curve_info_list=None,   # overriding sub_curve_info
                    ):
        # print('x.shape=', x.shape, 'y.shape=', y.shape, 'data.shape=', data.shape)
        self.dialog = dialog
        self.data = data
        self.x  = x
        self.y  = y

        self.epos = epos
        self.ecurve = ecurve
        self.el_color = el_color
        self.restrict_list = restrict_list
        self.toggle_info = toggle_info
        self.using_xray_conc = get_setting('use_xray_conc') == 1
        self.using_normal_conc = using_normal_conc()

        Tk.Frame.__init__(self, parent)
        self.build_body(titles, disable_x_selection, averaged_ridge, changeable_yscale, sub_curve_info, pre_recog, bridge_slice, sub_curve_info_list)

    def get_frame_classes(self):
        return OutlineViewer3dFrame, OutlineViewer2dFrame

    def build_body(self, titles, disable_x_selection, averaged_ridge, changeable_yscale, sub_curve_info, pre_recog, bridge_slice, sub_curve_info_list=None):
        class_3d, class_2d = self.get_frame_classes()

        body_frame = Tk.Frame(self)
        body_frame.pack(padx=10)

        W_frame = Tk.Frame(body_frame)
        W_frame.pack(side=Tk.LEFT, padx=5)
        self.W_frame = W_frame
        Wupper_frame = Tk.Frame(W_frame)
        Wupper_frame.pack()
        Wlower_frame = Tk.Frame(W_frame)
        Wlower_frame.pack()

        state = Tk.DISABLED if self.using_xray_conc else Tk.NORMAL
        self.toggle_btn = Tk.Button(Wlower_frame, text=self.toggle_info[0], command=self.toggle_info[1], state=state)
        self.toggle_btn.pack(side=Tk.RIGHT, padx=10)

        E_frame = Tk.Frame(body_frame)
        E_frame.pack(side=Tk.RIGHT, padx=5)
        Eupper_frame = Tk.Frame(E_frame)
        Eupper_frame.pack()
        Elower_frame = Tk.Frame(E_frame)
        Elower_frame.pack()

        if is_low_resolution():
            figsize1 = ( 8, 7 )
            figsize2 = ( 8, 3 )
        else:
            figsize1 = ( 10, 8 )
            figsize2 = ( 10, 4 )

        ppi = self.ecurve.primary_peak_i
        if averaged_ridge:
            pp_ridge_y = self.get_averaged_ridge(ppi)
        else:
            pp_ridge_y = self.data[:,ppi]

        ridge_color = 'green'
        curve_info = [(self.epos, self.ecurve.y, self.el_color), (ppi, pp_ridge_y, ridge_color)]

        if sub_curve_info_list is None:
            if sub_curve_info is not None:
                sub_curve_info_list = [sub_curve_info]      # assure backward compatibility

        if sub_curve_info_list is not None:
            curve_info += sub_curve_info_list

        canvas1 = class_3d(Wupper_frame, self, figsize1, titles[0], self.x, self.y, self.data, curve_info)
        canvas1.pack()

        e_restrict = None if self.restrict_list is None else self.restrict_list[0]
        canvas2 = class_2d(Eupper_frame, self, figsize2, titles[1], self.y, self.ecurve.y, self.data, self.el_color, ppi, ridge_color, e_restrict,
                                        sub_curve_info=sub_curve_info, pre_recog=pre_recog, ecurve=self.ecurve,
                                        toolbar=False)  # TODO: toolbar=True
        canvas2.pack()

        opos_sub = None if sub_curve_info is None else sub_curve_info[0]
        r_restrict = None if self.restrict_list is None else self.restrict_list[1]
        canvas3 = class_2d(Elower_frame, self, figsize2, titles[2], self.x, pp_ridge_y, self.data, ridge_color, self.epos, self.el_color, r_restrict,
                                opos_sub=opos_sub,
                                disable_selection=disable_x_selection,
                                changeable_yscale=changeable_yscale,
                                bridge_slice=bridge_slice,
                                toolbar=True, val_label=True)
        canvas3.pack()
        self.canvases = []
        for k, c in enumerate([canvas1, canvas2, canvas3]):
            if self.using_normal_conc or k == 1:
                c.mpl_canvas.draw()
            self.canvases.append(c)

    def close(self):
        for canvas in self.canvases:
            canvas.close()

    def get_averaged_ridge(self, ppi):
        start = max(0, ppi - 2)
        stop  = min(self.data.shape[1], ppi + 2)
        y = np.average(self.data[:, start:stop], axis=1)
        return y

    def get_info_list(self):
        info_list = []
        for c in self.canvases[1:]:
            info_list.append(c.get_restrict_info())
        return info_list

    def forget_left_frame(self):
        self.W_frame.pack_forget()

    def pack_left_frame(self):
        self.W_frame.pack()

    def disable_toggle_button(self):
        self.toggle_btn.config(state=Tk.DISABLED)

class OutlineViewerDialog(Dialog):
    def __init__(self, parent, serial_data):
        self.logger = logging.getLogger( __name__ )
        self.serial_data = serial_data
        self.using_normal_conc = using_normal_conc()
        self.using_xray_conc = get_setting('use_xray_conc') == 1
        self.xray_only = False
        Dialog.__init__(self, parent, "Outline Viewer", visible=False)

    def get_frame_class(self):
        return OutlineViewerFrame

    def get_bridge_slice(self):
        return self.serial_data.xray_slice

    def body(self, frame):
        frame_class = self.get_frame_class()

        serial_data = self.serial_data
        self.frames = []

        titles = ["Xray Scattering Data", "Elution Curve in Xray", "Xray Scattering Curve"]
        x   = serial_data.qvector
        y   = serial_data.jvector
        xray_slice = self.serial_data.xray_slice
        epos = (xray_slice.start + xray_slice.stop)//2
        ecurve  = serial_data.get_xray_curve()
        data = serial_data.intensity_array[:, :, 1].T
        el_color = 'orange'
        xr_restrict_list = get_setting('xr_restrict_list')
        bridge_slice = self.get_bridge_slice()

        f = frame_class(frame, self, titles, x, y, data, epos, ecurve, el_color, xr_restrict_list,
                            ["Show UV", self.toggle], changeable_yscale=True, averaged_ridge=False,
                            bridge_slice=bridge_slice)
        f.pack()
        self.frames.append(f)

        if self.xray_only:
            f = Tk.Frame(frame)     # dummy (never be apparent)
        else:
            titles = ["UV Absorbance Data", "Elution Curve in UV", "UV Absorbance Curve"]
            absorbance = serial_data.absorbance
            data = absorbance.data
            x   = absorbance.wl_vector
            y   = np.arange(data.shape[1])
            absorbance = serial_data.absorbance
            ecurve  = absorbance.a_curve
            epos    = absorbance.index    # this line must be executed after get_uv_curve()
            el_color = 'blue'
            uv_restrict_list = get_setting('uv_restrict_list')

            sub_curve_info = None
            sub_curve_info_list = None
            if self.using_normal_conc:
                # task: make it easier to give these specifications
                from molass_legacy.UV.PlainCurveUtils import get_flat_info
                w, i, ev = get_flat_info(self.serial_data)
                sub_curve_info_list = [ (absorbance.index_sub, absorbance.a_vector_sub, 'blue', ':'),
                                        (i, ev, 'cyan', ':', True),
                                        ]

            x_axis_ends=serial_data.absorbance.get_wave_len_ends()
            print('x_axis_ends=', x_axis_ends)
            f = frame_class(frame, self, titles, x, y, data, epos, ecurve, el_color, uv_restrict_list,
                                ["Show Xray", self.toggle], disable_x_selection=True,
                                sub_curve_info=sub_curve_info,
                                pre_recog=serial_data.pre_recog,
                                x_axis_ends=x_axis_ends,
                                sub_curve_info_list=sub_curve_info_list)
        f.pack()
        self.frames.append(f)

        self.current = 0
        self.frames[1 - self.current].pack_forget()

    def get_current_frame(self):
        return self.frames[self.current]

    def get_opposite_frame(self):
        return self.frames[1 - self.current]

    def show(self):
        self._show()

    def toggle(self):
        self.current = 1 - self.current
        self.frames[1 - self.current].pack_forget()
        self.frames[self.current].pack()

    def cancel(self):
        for frame in self.frames:
            frame.close()

        Dialog.cancel(self)



# coding: utf-8
"""
    AutorgKekAdapter.py

    Copyright (c) 2017-2021, SAXS Team, KEK-PF
"""
import numpy as np
from SimpleGuinier      import SimpleGuinier
from Result             import Result
from ErrorResult        import ErrorResult
from WLS_Rg             import WLS_Rg
from Quality            import STDEV_SCORE_SCALE
from LightObjects       import AnyObject, LightIntensity
from scipy.interpolate  import UnivariateSpline
from molass_legacy._MOLASS.SerialSettings     import get_setting

DEBUG   = False

def get_new_factor_weights():
    return get_setting( 'quality_weighting' )

def new_q_rg_score_func( qrg ):
    if qrg <= 0.2: return 1
    if qrg >= 1.2: return 0
    return ( 1.2 - qrg ) / 1.0

class AdapterQualiy:
    def __init__( self, sg, wls ):
        self.basic_condition    = 0
        self.stdev_ratio    = stdev_ratio = wls.sigmaRg / wls.Rg
        self.basic_quality  = basic_quality = sg.basic_quality
        positive_ratio      = self.compute_positive_ratio( sg.y_ )
        self.positive_score = 2 * ( max( 0.5, positive_ratio ) - 0.5 )
        self.fit_consistency    = sg.end_consistency
        self.fit_consistency_pure   = 0
        stdev_score         = np.exp( -stdev_ratio*STDEV_SCORE_SCALE ) if stdev_ratio > 0 else 0
        self.stdev_score    = stdev_score * basic_quality
        q_rg_score_         = new_q_rg_score_func( sg.x[sg.guinier_start]*sg.Rg )
        self.q_rg_score     = q_rg_score_ * basic_quality
        self.fit_score      = 0
        self.update_quality()

    def compute_positive_ratio( self, I ):
        N = I.shape[0]

        y05, y95 = np.percentile( I, [ 5, 95 ] )
        if y05 > 0:
            if DEBUG: print( 'y05, y95=', y05, y95 )
            dy = ( y95 - y05 )/( N - 1 )
            weights = np.arange( y95, y05 - dy/10, -dy )
            if DEBUG: print( 'N=', N, 'weights.shape=', weights.shape )
        else:
            weights = np.ones( ( N, ) )

        # print( 'weights.shape=', weights.shape, 'N=', N )
        assert( weights.shape[0] == N )

        weights /= np.sum( weights )
        ratio = np.sum( weights[ I > 0 ] )
        if DEBUG: print( 'positive_ratio=', ratio )
        return ratio

    def get_raw_factors( self ):
        return  [   self.basic_quality,
                    self.positive_score,
                    self.fit_consistency,
                    self.fit_consistency,
                    self.stdev_score,
                    self.q_rg_score,
                    self.fit_score,
                ]

    def get_factors( self ):
        weights = get_new_factor_weights()
        return  [   weights[0] * self.basic_quality,
                    weights[1] * self.positive_score,
                    0,
                    weights[2] * self.fit_consistency,
                    weights[3] * self.stdev_score,
                    weights[4] * self.q_rg_score,
                ]

    def get_factors_with_fit_score( self ):
        return self.get_factors() + [ self.fit_score ]

    def update_quality( self ):
        self.quality =  np.sum( self.get_factors() )
        if np.isnan( self.quality ):
            # TODO: remove this case
            print( 'quality factors=', self.get_factors() )
            import logging
            logger = logging.getLogger( __name__ )
            logger.warning( 'quality is set to 0; quality factors=' + str(self.get_factors()) )
            self.quality = 0

        if self.quality <= 0.3:
            self.quality_signal = 'R'
        elif self.quality >= 0.7:
            self.quality_signal = 'G'
        else:
            self.quality_signal = 'Y'

class AutorgKekAdapter:
    def __init__( self, file, guinier=None ):
        if guinier is None:
            self.sg = SimpleGuinier( file )
        else:
            self.sg = guinier
        self.intensity  = LightIntensity( AnyObject( orig_array=self.sg.data ) )

    def run( self, robust=None, optimize=None, fit_result=False, write_log=False ):
        ret = self.run_impl( robust, optimize, fit_result  )

        if write_log:
            from molass_legacy.Test.Tester import write_to_tester_log
            sg = self.sg
            write_to_tester_log( 'guinier result: type=%s(%s), Rg=(%s, %s), basic_quality=%g, score_vector=%s\n'
                        % ( sg.peak_type, str(sg.try_ret), str(ret.Rg), str(sg.Rg), sg.basic_quality, str(sg.score_vector) ) )

        return ret

    def run_impl( self, robust=None, optimize=None, fit_result=False ):
        if self.sg.Rg is None:
            # raise RuntimeError()
            return ErrorResult()

        slice_ = slice( self.sg.guinier_start, self.sg.guinier_stop )
        x_  = self.sg.x2[slice_]
        y_  = self.sg.log_y[slice_]
        e_  = self.sg.log_y_err[slice_]

        w_ = 1/e_**2
        wls = WLS_Rg( x_, y_, w_ )
        if wls.Rg is None:
            w_ = np.ones( len(x_) )
            wls = WLS_Rg( x_, y_, w_ )
            if wls.Rg is None:
                # raise RuntimeError()
                return ErrorResult()

        quality_obj = AdapterQualiy( self.sg, wls )
        if fit_result:
            spline = UnivariateSpline( x_, y_, s=0, ext=3 )
            fit = AnyObject( Rg=99, I0=0, degree=1, q1=0.1, model=spline, result=AnyObject( aic=-1, bic=-1 ) )
        else:
            fit = AnyObject( Rg=0, I0=0, degree=0, result=AnyObject( aic=-1, bic=-1 ) )

        fr_ = self.sg.guinier_start
        to_ = self.sg.guinier_stop - 1

        return Result(
            type = 'A',     # Adapter
            Rg          = wls.Rg,
            Rg_         = wls.Rg,
            Rg_stdev    = wls.sigmaRg,
            I0          = wls.I0,
            I0_         = wls.I0,
            I0_stdev    = wls.sigmaI0,
            From        = fr_,
            To          = to_,
            min_q       = self.sg.min_q,
            max_q       = self.sg.max_q,
            a   = wls.B,
            b   = wls.A,
            fit         = fit,
            min_qRg     = self.sg.min_qRg,
            max_qRg     = self.sg.max_qRg,
            bico_mono_ratio = 1,
            head_trend  = 0,
            IpI         = 0,
            bicomponent = 0,
            result_type = 0,
            quality_object  = quality_obj,
            basic_quality = quality_obj.basic_quality,
            Quality     = quality_obj.quality,
            rg_ratio_for_better_peak = self.sg.rg_ratio_for_better_peak,
            head_gap_ratio = self.sg.head_gap_ratio,
            )



"""
    V2PropOptimizer.VariedPropAnalysis.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import re
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar, get_color
from ScrolledFrame import ScrolledFrame
from molass_legacy._MOLASS.SerialSettings import get_setting
from DataUtils import get_in_folder
from molass_legacy._MOLASS.Version import is_developing_version

VARYING_INUIT_PARAMS = False

class VariedPropAnalysis(Dialog):
    def __init__(self, parent, js_canvas, modelname, num_variations=20):
        self.parent = parent
        self.js_canvas = js_canvas
        self.modelname = modelname
        self.num_variations = num_variations
        Dialog.__init__(self, parent, "Varied Proportion Analysis", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        self.scrolled_frame = ScrolledFrame(body_frame)
        self.scrolled_frame.pack(anchor=Tk.N)
        cframe = self.scrolled_frame.interior

        self.ncols = 5
        self.nrows = self.num_variations//self.ncols
        height = 2.5 * self.nrows
        fig, axes = plt.subplots(nrows=self.nrows, ncols=self.ncols, figsize=(20,height))
        self.draw_varied_proportions(fig, axes)
        fig.tight_layout()

        self.fig = fig
        self.axes = axes
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.popup_menu = None
        self.mpl_canvas.mpl_connect('button_press_event', self.on_figure_click)

    def buttonbox(self):
        # task: pack the close button so that it won't hide
        box = Tk.Frame(self)
        box.pack(fill=Tk.X)

        tframe = Tk.Frame(box)
        tframe.pack(side=Tk.LEFT, padx=20, pady=10)
        bframe = Tk.Frame(box)
        bframe.pack(side=Tk.RIGHT, padx=20, pady=10)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        w = Tk.Button(bframe, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=20)

        w = Tk.Button(bframe, text="RDR Chart", width=10, command=self.show_rdr_chart)
        w.pack(side=Tk.LEFT, padx=20)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)
        self.set_geometry()

    def set_geometry(self):
        self.update()
        canvas_width = int(self.mpl_canvas_widget.cget( 'width' ))
        canvas_height = int(self.mpl_canvas_widget.cget( 'height' ))
        print("canvas_height=", canvas_height)
        height = min(900, canvas_height+200)
        margin_width = 40
        wxh = '%dx%d' % (canvas_width + margin_width, height)
        geometry = self.geometry()
        new_geometry = re.sub( r'(\d+x\d+)(.+)', lambda m: wxh + m.group(2), geometry)
        self.geometry(new_geometry)

    def draw_varied_proportions(self, fig, axes, devel=True):
        if devel:
            from importlib import reload
            import V2PropOptimizer.PropOptimizer
            reload(V2PropOptimizer.PropOptimizer)
            import V2PropOptimizer.PropOptimizerUtils
            reload(V2PropOptimizer.PropOptimizerUtils)
            import ElutionCurve
            reload(ElutionCurve)
        from V2PropOptimizer.PropOptimizer import compute_range_rgs
        from V2PropOptimizer.PropOptimizerUtils import PropOptimizer
        from molass_legacy._MOLASS.SerialSettings import set_setting
        from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve

        self.optimizer = optimizer = self.js_canvas.fullopt
        xr_curve = optimizer.xr_curve
        x = xr_curve.x
        y = xr_curve.y

        self.prop_optimizer = prop_optimizer = PropOptimizer(self.modelname, x, y)
        init_peaks = prop_optimizer.get_init_params()
        props = prop_optimizer.compute_props(init_peaks)
        print("props=", props)

        self.pv = np.linspace(props[0], 0.2, self.num_variations)

        fv_list = []
        rdr_list = []
        peaks_list = []

        # set_setting('local_debug', True)
        # ecurve = ElutionCurve(y, x=x)     # there still is a bug
        ecurve = ElutionCurve(y)
        self.paired_ranges = paired_ranges = ecurve.get_default_paired_ranges()
        # set_setting('local_debug', False)
        print("paired_ranges=", paired_ranges)

        qv = optimizer.qvector
        D = optimizer.xrD
        E = optimizer.xrE

        minRDR = None
        n = None

        fig.suptitle("Varied Proportion Avalysis of %s using %s" % (get_in_folder(), self.modelname), fontsize=20)

        for i, p in enumerate(self.pv):
            if i != 17:
                # continue
                pass

            print([i], "optimizing with p=%.3g" % p)
            j, k = divmod(i,5)
            ax = axes[j,k]
            ax.plot(x, y, color="orange")
            opt_ret = prop_optimizer.optimize((p, 1 - p), init_params=init_peaks)
            opt_peaks = opt_ret.x.reshape(init_peaks.shape)
            fv_list.append(opt_ret.fun)
            peaks_list.append(opt_peaks)

            cy_list = prop_optimizer.compute_cy_list(opt_peaks)

            if VARYING_INUIT_PARAMS:
                init_peaks = opt_peaks

            for cy in cy_list:
                ax.plot(x, cy, ":")
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red")
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            tx = xmin*0.9 + xmax*0.1
            ty = ymin*0.2 + ymax*0.8
            ax.text(tx, ty, "p=%.3g" % p, fontsize=20, alpha=0.3)

            C = np.array(cy_list)
            rgs = compute_range_rgs(qv, D, E, paired_ranges, C)

            rdr = abs(rgs[0] - rgs[1])*2/(rgs[0] + rgs[1])
            rdr_list.append(rdr)
            if minRDR is None or rdr < minRDR:
                minRDR = rdr
                n = i
            ty = ymin*0.4 + ymax*0.6
            ax.text(tx, ty, "Rgs=%.3g, %.3g" % tuple(rgs), fontsize=20, alpha=0.3)

        if n is not None:
            j, k = divmod(n,5)
            ax = axes[j,k]
            ax.patch.set_facecolor('green')
            ax.patch.set_alpha(0.1)

        self.fv = np.array(fv_list)
        self.rdr = np.array(rdr_list)
        self.peaks_list = peaks_list

    def show_fv_chart(self):
        indeces = [str(i) for i in range(self.num_variations)]
        with dplt.Dp(window_title='FV Chart', ok_only=True, ok_text="Close"):
            fig, ax = dplt.subplots()
            ax.set_title("FV Chart", fontsize=16)
            ax.bar(indeces, self.fv)
            fig.tight_layout()
            dplt.show()

    def show_rdr_chart(self, devel=True):
        if devel:
            from importlib import reload
            import V2PropOptimizer.RdrChart
            reload(V2PropOptimizer.RdrChart)
        from .RdrChart import draw_rdr_chart
        draw_rdr_chart("RDR_AD Chart with %s" % self.modelname, self.rdr)

    def on_figure_click(self, event):
        if event.button == 3:
            axes = self.axes
            for i in range(axes.shape[0]):
                for j in range(axes.shape[1]):
                    if event.inaxes == axes[i,j]:
                        self.current_index = (i,j)
                        self.current_event = event
                        self.show_popup_menu(event)
                        break

    def show_popup_menu(self, event):
        from molass_legacy.KekLib.TkUtils import split_geometry
        self.create_popup_menu(event)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu.post(cx + int(event.x), cy + h - int(event.y))

    def create_popup_menu(self, event):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu(self, tearoff=0 )
            self.popup_menu.add_command(label='Observe it as V1 Preview', command=self.observe_it_as_v1_preview)
            if is_developing_version():
                self.popup_menu.add_command(label='Optimize Proportion with EGH', command=lambda: self.optimize_proportion('EGH'))
                self.popup_menu.add_command(label='Optimize Proportion with EDM', command=lambda: self.optimize_proportion('EDM'))
                self.popup_menu.add_command(label='Optimize Proportion with STC', command=lambda: self.optimize_proportion('STC'))

    def optimize_proportion(self, name, devel=True):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        if devel:
            from importlib import reload
            import V2PropOptimizer.PropOptimizerDialogs
            reload(V2PropOptimizer.PropOptimizerDialogs)
        from V2PropOptimizer.PropOptimizerDialogs import show_optimizer_dialog

        i, j = self.current_index
        k = self.ncols*i + j
        p = self.pv[k]
        print("optimize_proportion: model=%s pv[%d]=%.3g" % (name, k, p))

        yn = MessageBox.askyesno("Proportion Optimization Confirmation",
            "Optimizing the proportion with model=%s, init_proportion=%.3g.\n"
            "Ok?" % (name, p),
            parent=self)

        if yn:
            self.after(500, self.cancel)
            show_optimizer_dialog(self.parent, name, p, self.optimizer)

    def observe_it_as_v1_preview(self, devel=True):
        if devel:
            from importlib import reload
            import V2PropOptimizer.V1PreviewAdapter
            reload(V2PropOptimizer.V1PreviewAdapter)
        from .V1PreviewAdapter import observe_it_as_v1_preview
        observe_it_as_v1_preview(self)

    def get_current_peaks(self):
        i, j = self.current_index
        k = self.ncols*i + j
        return self.peaks_list[k]



# coding: utf-8
"""
    VoxelFFT.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Polygon, Circle
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import get_color
from OurManim import manim_init, use_default_style, Animation, Collection, TextGroup, Parallelogram, rotation, angle
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from .VoxelUtils import draw_voxel

class VoxelFFT(Dialog):
    def __init__(self, parent=None):
        if parent is None:
            from SerialTestUtils import get_tk_root
            parent = get_tk_root()
        self.parent = parent
        # manim_init()
        Dialog.__init__(self, parent, "SVD Tutorial", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig = fig = plt.figure(figsize=(16,8))
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, cframe)
        self.toolbar.update()
        ax1 = fig.add_subplot(121, projection='3d')
        ax2 = fig.add_subplot(122)
        self.axes = ax1, ax2
        self.draw()
        self.fig.tight_layout()
        self.mpl_canvas.draw()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Save", width=10, command=self.save)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

    def draw(self):
        ax, _ = self.axes
        # tensor = np.random.rand(3, 3, 3)
        tensor = np.ones((3, 3, 3))
        print(tensor)
        draw_voxel(ax, tensor)

        ax.set_xlim(-10, 10)
        ax.set_ylim(0, 20)
        ax.set_zlim(-10, 10)

    def save(self):
        pass



"""

    Models.EMG.py

    Copyright (c) 2017-2024, SAXS Team, KEK-PF

"""
import numpy                as np
import matplotlib.pyplot    as plt
from matplotlib.patches     import Polygon
from bisect                 import bisect_right
from scipy.interpolate      import UnivariateSpline
from scipy.optimize import minimize
from molass_legacy.KekLib.Affine import Affine
from molass_legacy.Peaks.ElutionModels import emg, VERY_SMALL_VALUE, LARGE_VALUE
from molass_legacy._MOLASS.SerialSettings import get_setting
from .ElutionModelUtils import x_from_height_ratio_impl
from .ModelUtils import _get_model_name

DEVEL = False
if DEVEL:
    from importlib import reload
    import Models.Tentative
    reload(Models.Tentative)
from .Tentative import Model

# task: these limits should be eventually superseded by TAU_BOUND_RATIO
EMG_TAU_LOWER_LIMIT = -50
EMG_TAU_UPPER_LIMIT = 50

def emg_x_from_height_ratio(alpha, mu, sigma, tau):
    n = 100
    hp = mu + tau
    x = np.linspace(hp-4*sigma, hp+5*sigma, n)      # 5*sigma is required for AhRR 
    y = emg(x, 1, mu, sigma, tau)
    hn = np.argmax(y)
    i = bisect_right(y[0:hn], alpha)
    j = bisect_right(-y[hn:], -alpha)

    if False:
        import molass_legacy.KekLib.DebugPlot as dplt
        print('i, hn, jn+j=', i, hn, hn+j)
        fig, ax = dplt.subplots()
        ax.plot(x, y)
        ax.plot(x[hn], y[hn], 'o', color='red')
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        try:
            for k in [i, hn+j]:
                ax.plot([x[k], x[k]], [ymin, ymax], ':', color='gray')
        except:
            pass
        dplt.show()

    def get_x0(k):
        # linear approximation
        if k == 0:
            return x[0]
        if k == len(x):
            return x[-1]

        x1 = x[k-1]
        x2 = x[k]
        y1 = y[k-1]
        y2 = y[k]
        return x1 - (x2-x1)/(y2-y1)*(y1-alpha)

    xi = get_x0(i)
    xj = get_x0(hn+j)
    return xi, xj

def emg_x_demo_plot():
    from molass_legacy.KekLib.OurMatplotlib import get_default_colors

    colors = get_default_colors()

    x = np.linspace( 0, 20, 200 )

    h       = 1
    mu      = 10
    sigma   = 1
    tau_list = [ 0, 0.5, 1, 1.5, 2, 3 ]

    fig = plt.figure( figsize=(10, 6) )
    ax  = fig.add_subplot(111)
    ax.set_title( "Extension of EMG for non-positive values of tau", fontsize=16 )

    for k, tau in enumerate( tau_list[1:] ):
        ax.plot( x, emg( x, h, mu, sigma, tau ), ':', color=colors[k+1], label='tau=%g' % ( tau ) )

    for k, tau in enumerate( tau_list ):
        ax.plot( x, emg( x, h, mu, sigma, -tau ), color=colors[k], label='tau=%g' % ( -tau ) )

    ax.legend(fontsize=16)

    plt.tight_layout()
    plt.show()

def emg_guess( x, y ):
    n   = np.argmax( y )
    h   = y[n]

    """
        https://en.wikipedia.org/wiki/Skewness
    """
    v   = np.sum(y)
    m   = np.sum( x*y )/ v
    m2  = np.sum( (x-m)**2 * y )/ v
    m3  = np.sum( (x-m)**3 * y )/ v
    s   = np.sqrt( m2 )
    sk  = m3 / np.power( m2, 3/2 )

    if sk > VERY_SMALL_VALUE:
        tau     = s * np.power( sk/2, 1/3 )
        mu      = m - tau
        sk_c    = 1 - np.power( sk/2, 2/3 )
        # TODO: sk_c <= 0 for OA_Ald
        sigma   = np.sqrt( s**2 * sk_c ) if sk_c > VERY_SMALL_VALUE else s
    elif sk < -VERY_SMALL_VALUE:
        tau     = -s * np.power( -sk/2, 1/3 )
        mu      = m - tau
        sk_c    = 1 - np.power( -sk/2, 2/3 )
        # TODO: sk_c <= 0
        sigma   = np.sqrt( s**2 * sk_c ) if sk_c > VERY_SMALL_VALUE else s
    else:
        tau     = 0
        mu      = m
        sigma   = s

    if False:
        print( 'm=', m, 'm2=', m2, 'm3=', m3, 's=', s, 'sk=', sk )
        print( 'h=', h, 'mu=', mu, 'sigma=', sigma, 'tau=', tau )

    if False:
        fig = dplt.figure()
        ax  = fig.add_subplot(111)
        ax.set_title( "emg_guess" )
        ax.plot( x, y, label='data' )
        ax.plot( x, emg( x, h, mu, sigma, tau ), label='guess' )
        ax.legend()
        dplt.show()

    return h, mu, sigma, tau

class EMG(Model):
    def __init__(self, **kwargs):
        Model.__init__(self, emg, **kwargs)

    def guess(self, y, x=None, negative=False, **kwargs):
        h, mu, sigma, tau = emg_guess( x, y )
        return np.array([h, mu, sigma, tau])

    def eval(self, params=None, **kwargs):
        ret = self.func(**self.make_funcargs(params, kwargs))
        ret[ np.logical_not( np.isfinite(ret) ) ] = LARGE_VALUE
        # TODO: this is not enough. still raises ValueError("The input contains nan values")
        return ret

    def get_name( self ):
        return _get_model_name( self.name )

    def x_from_height_ratio(self, ecurve, ratio, params):
        # mu, sigma, tau
        return x_from_height_ratio_impl(emg, ecurve, ratio, *params[1:])

    def get_peaktop_xy_bad(self, x, params):
        px = np.sum(params[[1,3]])      # mu, tau
        py = self.func(px, *params)
        return px, py

    def get_params_string(self, params):
        return 'h=%g, mu=%g, sigma=%g, tau=%g' % tuple(params)

    def guess_a_peak_with_prop(self, x, y, prop, debug=True):
        """
            This function is used to guess a modeled peak with a given area ratio.
        """
        h, mu, sigma, tau = emg_guess(x, y)
        total_area = np.sum(y)
        def objective(p, debug=False):
            sigma_, tau_  = p
            cy = emg(x, h, mu, sigma_, tau_)
            if debug:
                import molass_legacy.KekLib.DebugPlot as dplt
                with dplt.Dp():
                    fig, ax = dplt.subplots()
                    ax.set_title('sigma=%g, tau=%g' % (sigma_, tau_))
                    ax.plot(x, y, label='data')
                    ax.plot(x, cy, label='guess')
                    ax.legend()
                    fig.tight_layout()
                    dplt.show()
            area = np.sum(cy)
            return abs(area - total_area*prop)
        objective([sigma, tau], debug=debug)
        ret = minimize(objective, [sigma, tau], method='Nelder-Mead')
        sigma, tau = ret.x
        return np.array([h, mu, sigma, tau])

CONSTR_A_SIGMA_RATIO    = 1.5
CONSTR_TAU_SIGMA_RATIO  = 5

def emga_impl(x, h, mu, sigma, tau, a, raise_=False ):
    y   = emg( x, h, mu, sigma, tau )

    if abs(a) < VERY_SMALL_VALUE/max(VERY_SMALL_VALUE, h):
        return y

    xf  = mu - 5*sigma
    xt  = mu + 5*sigma

    src_points  = [ (xf, 0), (mu, h), (xt, 0) ]
    tgt_points  = [ (xf, 0), (mu + a, h ), (xt, 0) ]

    try:
        affine = Affine( src_points, tgt_points, raise_=raise_ )
        x_, y_ = affine.transform( x, y )
        if len(x_) > 1:
            spline = UnivariateSpline( x_, y_, s=0, ext=3 )
            ret = spline( x )
        else:
            ret = y_
    except:
        ret = np.ones(len(x)) * np.nan

    return ret

def emga(x, h, mu, sigma, tau, a):
    if np.isscalar(x):
        x = np.array([x])
        return emga_impl(x, h, mu, sigma, tau, a)[0]
    else:
        return emga_impl(x, h, mu, sigma, tau, a)

def emg_no_affine(x, h=1.0, mu=0.0, sigma=1.0, tau=1.0, a=0):
    return emg(x, h, mu, sigma, tau)

ALLOWANCE_MIN_EQ_MAX = 1e-12

def emga_guess_impl(self, y, x, negative, **kwargs):
    h, mu, sigma, tau = emg_guess( x, y )
    return np.array([h, mu, sigma, tau, 0])

class EMGA(EMG):
    def __init__(self, **kwargs):
        func = emga if get_setting('enable_affine_tran') else emg_no_affine
        Model.__init__(self, func, **kwargs)

    def guess(self, y, x=None, negative=False, **kwargs):
        return emga_guess_impl( self, y, x, negative, **kwargs )

    def eval(self, params=None, x=None):
        ret = self.func(x, *params)
        ret[ np.logical_not( np.isfinite(ret) ) ] = LARGE_VALUE
        # TODO: this is not enough. still raises ValueError("The input contains nan values")
        return ret

    def get_name( self ):
        return _get_model_name( self.name )

    def get_param_hints( self, pkey ):
        assert pkey == 'tau'
        return EMG_TAU_LOWER_LIMIT, EMG_TAU_UPPER_LIMIT

    def get_def_expr( self, params ):
        return "emga( h=%.3g, mu=%.3g, sigma=%.3g, tau=%.3g, a=%.3g )" % (params['h'], params['mu'], params['sigma'], params['tau'], params['a'])

    def x_from_height_ratio(self, ecurve, ratio, params):
        # mu, sigma, tau, a
        return x_from_height_ratio_impl(emga, ecurve, ratio, *params[1:])

    def get_params_string(self, params):
        return 'h=%g, mu=%g, sigma=%g, tau=%g, a=%g' % tuple(params)
    
    def guess_a_peak_with_prop(self, x, y, prop, debug=True):
        """
            This function is used to guess a modeled peak with a given area ratio,
            which has been generated (after the implementation of the same method of EMG)
            completely by Cody AI.
        """
        h, mu, sigma, tau = emg_guess(x, y)
        total_area = np.sum(y)
        m = np.argmax(y)
        px = x[m]
        py = y[m]
        def objective(p, debug=False):
            h_, mu_, sigma_, tau_, a_ = p
            cy = emga(x, h_, mu_, sigma_, tau_, a_)
            k = np.argmax(cy)
            px_ = x[k]
            py_ = cy[k]
            if debug:
                import molass_legacy.KekLib.DebugPlot as dplt
                with dplt.Dp():
                    fig, ax = dplt.subplots()
                    ax.set_title('sigma=%g, tau=%g, a=%g' % (sigma_, tau_, a_))
                    ax.plot(x, y, label='data')
                    ax.plot(x, cy, label='guess')
                    ax.legend()
                    fig.tight_layout()
                    dplt.show()
            area = np.sum(cy)
            return max(np.log10((area - total_area*prop)**2), 5*np.log10((px_ - px)**2 + (py_ - py)**2))

        objective([h, mu, sigma, tau, 0], debug=debug)
        ret = minimize(objective, [h, mu, sigma, tau, 0], method='Nelder-Mead')
        h, mu, sigma, tau, a = ret.x
        return np.array([h, mu, sigma, tau, a])

def emga_demo_plot(with_xy_map=False):
    x   = np.linspace( 0, 300, 300 )
    if with_xy_map:
        fig = plt.figure( figsize=(24,6) )
        # print( fig.get_size_inches() )
        ax  = fig.add_subplot(131)
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)
    else:
        fig = plt.figure()
        # print( fig.get_size_inches() )
        ax  = fig.add_subplot(111)

    ax.set_title( 'EMGA: EMG with Affine transformation' )
    ax.set_ylim( -0.1, 1.2 )
    # ax.plot( x, emg( x, 1, 150, 30, 30 ) )
    h       = 1
    mu      = 150
    sigma   = 30
    tau     = 10
    a       = 40
    y1  = emga( x, h, mu, sigma, tau, 0 )
    ax.plot( x, y1 )
    y2  = emga( x, h, mu, sigma, tau, a )
    ax.plot( x, y2 )
    xf  = mu - 5*sigma
    xt  = mu + 5*sigma
    src_points = [ (xf, 0), (mu,h), (xt, 0) ]
    tgt_points = [ (xf, 0), (mu+a,h), (xt, 0) ]
    src_polygon = Polygon( src_points, alpha=0.1 )
    ax.add_patch(src_polygon)
    tgt_polygon = Polygon( tgt_points, alpha=0.1, fc='green' )
    ax.add_patch(tgt_polygon)

    colors  = [ 'orange', 'red', 'red', 'yellow' ]
    points  = [ ( xf, 0), (mu, h), (mu+a, h), (xt, 0) ]
    for k, p in enumerate(points):
        ax.plot( p[0], p[1], 'o', color=colors[k] )

    ax.annotate("(μ, h)", xy=(mu, h), xytext=(mu, h + 0.1), ha='center', arrowprops=dict( arrowstyle="-", color='k' ) )
    ax.annotate("(μ + a, h)", xy=(mu+a, h), xytext=(mu+a, h + 0.1), ha='center', arrowprops=dict( arrowstyle="-", color='k' ) )
    ax.annotate("", xy=(mu+a, h), xytext=(mu, h), size=20, arrowprops=dict( arrowstyle="->", linewidth=3, color='k' ))

    ax.annotate( "(μ - 5*σ, 0)", xy=(xf, 0), xytext=(xf, 0.1), ha='center', arrowprops=dict( arrowstyle="-", color='k' ) )
    ax.annotate( "(μ + 5*σ, 0)", xy=(xt, 0), xytext=(xt, 0.1), ha='center', arrowprops=dict( arrowstyle="-", color='k' ) )

    if with_xy_map:
        ax2.set_title( "Mapping of x in the Affine transformation" )
        ax2.set_xlabel( "Elution No." )
        ax2.set_ylabel( "Transformed Elution No." )
        ax3.set_title( "Mapping of y in the Affine transformation" )
        ax3.set_xlabel( "Intensity of EMG" )
        ax3.set_ylabel( "Intensity of EMGA" )
        affine = Affine( src_points, tgt_points )
        x_, y_ = affine.transform( x, y1 )
        ax2.plot( x, x_, 'o', markersize=3 )
        ax3.plot( y1, y_, 'o', markersize=3 )

    fig.tight_layout()
    plt.show()


"""
module for common operations on discrete random variables (distributions)
"""


import numpy as np

def assert_shape_equal(x_var, y_var):
    """
    Assert shape equal
    """
    if x_var.shape != y_var.shape:
        raise AssertionError('Shape mismatch!')


class DiscreteRV:
    """ class for 1-d finite discrete RV
    """
    def __init__(self, w, x):
        """
        weights: probabilites masses
        atoms: atoms
        """
        self.weights = np.asarray(w)
        self.atoms = np.asarray(x)
        assert_shape_equal(self.weights, self.atoms)

    def __repr__(self):
        return "atom: %s\nwght: %s" % (self.atoms, self.weights)

    def moment(self, degree=1):
        """ Compute the moments of the input RV up to the given degree (start from first degree)
        Args:
        degree: int
        highest degree k

        Returns:
        array of moments from the first degree to degree k
        """
        moments = np.zeros(degree)
        monomial = np.ones(self.atoms.shape)

        for i in range(degree):
            monomial *= self.atoms
            moments[i] = np.dot(self.weights, monomial)

        return moments

    def dist_w1(self, another_rv):
        """
        Compute the W1 distance from another_rv
        """
        return wass(self, another_rv)


    def sample(self, num):
        """ draw n iid sample from U
        library: https://docs.scipy.org/doc/numpy/reference/routines.random.html
        alternative way: scipy.stats.rv_discrete(name='custm', values=(xk, pk))
        """
        ## for 1-d RV
        return np.random.choice(self.atoms, size=num, replace=True, p=self.weights)

        ## for U in higher dimensions
        # return U.x[np.random.choice(U.x.shape[0],size=n, replace=True, p=U.p)]



    def sample_noisy(self, num, sigma=1):
        """ draw n iid samples from model U+sigma Z. Default sigma=1.
        """
        return self.sample(num) + sigma*np.random.randn(num)

def wass(u_rv, v_rv):
    """ compute W1 distance between DiscreteRVs U and V
    """
    if len(u_rv.atoms) == 0 or len(v_rv.atoms) == 0:
        return 0.

    x_u, p_u = zip(*sorted(zip(u_rv.atoms, u_rv.weights)))
    x_v, p_v = zip(*sorted(zip(v_rv.atoms, v_rv.weights)))
    l_u, l_v, diff_cdf, dist, pre = 0, 0, 0., 0., 0.
    while l_u < len(x_u) or l_v < len(x_v):
        if l_v == len(x_v) or (l_u < len(x_u) and x_v[l_v] > x_u[l_u]):
            dist += abs(diff_cdf)*(x_u[l_u]-pre)
            pre = x_u[l_u]
            diff_cdf += p_u[l_u]
            l_u += 1
        else:
            dist += abs(diff_cdf)*(x_v[l_v]-pre)
            pre = x_v[l_v]
            diff_cdf -= p_v[l_v]
            l_v += 1

    return dist



# coding: utf-8
"""
    SdDebugger.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import inspect
from collections import OrderedDict
from DataDebugger import DataDebugger

class SdDebugger(DataDebugger):
    def __init__(self):
        DataDebugger.__init__(self, "sd-debugger-00")

    def save_info(self, sd):
        sub_folder = self.get_sub_folder()
        np.savetxt(sub_folder + '/q.dat', sd.intensity_array[0,:,0])
        np.savetxt(sub_folder + '/M.dat', sd.intensity_array[:,:,1].T)

        param_info = OrderedDict()
        param_info['sd_id_info'] = sd.get_id_info()

        with open(sub_folder + '/param_info.txt', 'w') as fh:
            frame = inspect.stack()[1]
            fh.write("%s(%d)\n" % (frame.filename, frame.lineno))
            fh.write(''.join(frame.code_context) + "----\n")
            fh.write(str(param_info) + '\n')



"""
    LanJorgensonEgh.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, root
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Peaks.ElutionModels import egh, e1, e2
from LPM import get_corrected
from .BoundControl import Penalties
from .ColumnConstants import Ti_LOWER
import molass_legacy.KekLib.DebugPlot as plt

ESTIMATE_INJECTION_TIME = True

def _pre_optimize_params(ecurve, xr_params, seccol_params, debug=False):
    x = ecurve.x
    y_ = ecurve.y

    # use corrected y
    y = get_corrected(y_)

    # Npc, rp, tI, t0, P, m
    Npc, tI = seccol_params[[0,2]]

    n = xr_params.shape[0]
    temp_xr_params = np.zeros((n,3))

    for k, (h, m, t) in enumerate(xr_params[:,[0,1,3]]):
        temp_xr_params[k,0] = h
        temp_xr_params[k,1] = m
        temp_xr_params[k,2] = t

    def objective(p):
        N_, T_ = p
        sqN = np.sqrt(N_)
        my = np.zeros(len(x))
        for k, (h, m, t) in enumerate(temp_xr_params):
            s = compute_sigma(T_, N_, m, t, init_sigma=xr_params[k,2])
            my += egh(x, h, m, s, t)
        return np.sum((my - y)**2)

    ret = minimize(objective, (Npc, tI))

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt

        with plt.Dp():
            fig, (ax0, ax1, ax2) = plt.subplots(ncols=3, figsize=(18,5))
            fig.suptitle("Lan-Jorgenson Adaption Proof", fontsize=20)
            ax0.set_title("Free EGH Decomposition", fontsize=16)
            ax1.set_title("Lan-Jorgenson Simple Conversion", fontsize=16)
            ax2.set_title("Lan-Jorgenson Optimized Adaption", fontsize=16)

            ax0.plot(x, y)
            for k, (h, m, s, t) in enumerate(xr_params):
                ax0.plot(x, egh(x, h, m, s, t), ":")

            def plot_params(ax, tI, Npc):
                ax.plot(x, y)
                for k, (h, m, t) in enumerate(temp_xr_params):
                    s = compute_sigma(tI, Npc, m, t, init_sigma=xr_params[k,2])
                    print([k], h, m, s, t)
                    ax.plot(x, egh(x, h, m, s, t), ":")
            plot_params(ax1, tI, Npc)
            plot_params(ax2, *ret.x)

            fig.tight_layout()
            plt.show()

    return temp_xr_params, ret.x

def lj_estimate_initial_params(ecurve, init_xr_params, seccol_params, debug=True):
    if debug:
        print("lj_estimate_initial_params")
        x = ecurve.x
        y = ecurve.y
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("lj_estimate_initial_params")
            ax.plot(x, y)
            cy_list = []
            for h, m, s, t in init_xr_params:
                cy = egh(x, h, m, s, t)
                cy_list.append(cy)
                ax.plot(x, cy, ":")
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red")
            fig.tight_layout()
            plt.show()

    temp_xr_params, (N_, T_) = _pre_optimize_params(ecurve, init_xr_params, seccol_params)
    lj_seccol_params = seccol_params.copy()
    lj_seccol_params[[0,2]] = N_, T_
    return temp_xr_params, lj_seccol_params

def lj_estimate_better(caller):
    print("lj_estimate_better")
    init_params_save = caller.init_params.copy()

    init_params = caller.compute_init_params()

    print("init_params_save=", init_params_save)
    print("init_params=", init_params)

    return

def compute_tau(tI, Np, mu, sigma):

    def equation(x):
        tau = x[0]
        tau_ = abs(tau)
        th = np.arctan2(tau_, sigma)
        M1 = mu + tau*e1(th) - tI
        M2 = (sigma**2 + sigma*tau_ + tau**2)*e2(th)
        return [M1**2 - M2*Np]

    ret = root(equation, [0])
    if ret.success:
        return ret.x[0]
    else:
        Penalties[0] += abs(equation(ret.x)[0])
        return 0

def compute_sigma(tI, Np, mu, tau, init_sigma=10):

    tau_ = abs(tau)

    def equation(x):
        sigma = x[0]
        th = np.arctan2(tau_, sigma)
        M1 = mu + tau*e1(th) - tI
        M2 = (sigma**2 + sigma*tau_ + tau**2)*e2(th)
        return [M1**2 - M2*Np]

    ret = root(equation, [init_sigma])

    return ret.x[0]

def convert_to_xr_params_lj(c_xr_params, tI, Np):
    nc = c_xr_params.shape[0]
    xr_params = np.zeros((nc, 4))
    sqrtNp = np.sqrt(Np)
    Penalties[0] = 0
    for k, (h, m, t) in enumerate(c_xr_params):
        xr_params[k,0] = h
        xr_params[k,1] = m
        init_sigma = (m - tI)/sqrtNp
        xr_params[k,2] = compute_sigma(tI, Np, m, t, init_sigma=init_sigma)
        xr_params[k,3] = t
    return xr_params

if __name__ == '__main__':
    pass



"""

    QuickAnalysis.AnalyzerDialogProxy.py

    Copyright (c) 2020-2025, SAXS Team, KEK-PF

"""
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

class AnalyzerDialogProxy:
    def __init__(self, parent, sd, mapper, judge_holder, correction_necessity):
        from molass_legacy.Extrapolation.PreviewButtonFrame import PreviewButtonFrame
        self.parent = parent
        self.doing_sec  = parent.use_mtd_conc == 0
        self.sd = sd
        self.mapper = mapper
        self.correction_necessity = correction_necessity
        use_elution_models=get_setting('use_elution_models')
        self.preview_frame = PreviewButtonFrame(parent, dialog=self, use_elution_models=use_elution_models, judge_holder=judge_holder)

    def make_range_info(self, *args):
        """
            concentration_datatype will be conveyed to later processes
            through the SerialSettings item set by preview.update_settings()
            independently from these args.
        """
        analysis_range_info = get_setting('analysis_range_info')
        return analysis_range_info.get_ranges()

    def get_decomp_info(self):
        # for compatibility with MctDecompDialog concerning Preview
        pdata, popts = get_setting('preview_params')
        return pdata.decomp_info

    def apply(self):
        self.parent.cfs_entry.update_conc_factors(self.sd, self.mapper)
        set_setting('averaged_data_folder', get_averaged_data_folder())
        set_setting('scattering_correction', int(self.correction_necessity))

        """
        TODO: this update is or can be a duplication. remove such inconvenience.
        """
        if self.doing_sec:
            self.preview_frame.update_settings()
        else:
            # 
            pass

def get_averaged_data_folder():
    # folder = get_setting( 'averaged_data_folder' )
    # if is_empty_val( folder ):
    folder = get_setting( 'analysis_folder' ) + '/averaged'
    return folder



"""
    ListingFrame.py

    adapted from DDList
        at
            Recipe11.4.Adding Drag and Drop Reordering to a Tkinter Listbox
            https://flylib.com/books/en/2.9.1.230/1/        

    Copyright (c) 2021-2022, Masatsuyo Takahashi, KEK-PF
"""
import tkinter as Tk

class ListingFrame(Tk.Frame):
    """ A frame with drag'n'drop reordering of entries. """
    def __init__(self, parent, **kw):
        self.numbering = kw.pop('numbering', False)
        self.debug = kw.pop('debug', False)
        Tk.Frame.__init__(self, parent, **kw)

        self.bind('<Button-1>', self.set_source)
        self.bind('<B1-Motion>', self.move_to_target)

        if self.numbering:
            self.number_labels = []
            self.widget_col = 1
        else:
            self.widget_col = 0
        self.pair_list = []
        self.h = None

    def insert(self, variable, widget):
        row=len(self.pair_list)
        if self.numbering:
            label = Tk.Label(self, text=str(row+1))
            label.grid(row=row, column=0, padx=5)
            self.number_labels.append(label)
        widget.grid(row=row, column=self.widget_col)
        widget.bind('<ButtonPress-1>', self.set_source)
        widget.bind('<B1-Motion>', self.move_widget_image)
        widget.bind('<ButtonRelease-1>', self.move_to_target)
        self.pair_list.append((variable, widget))

    def update_bind_geometry(self):
        self.update()
        self.width = self.winfo_reqwidth()
        self.height = self.winfo_reqheight()
        self.h = self.height/len(self.pair_list)
        if self.debug:
            print(self.width, self.height, self.h)

    def get_widget_index(self, event):
        rootx = self.winfo_rootx()
        rooty = self.winfo_rooty()
        x = event.x_root - rootx
        y = event.y_root - rooty
        i = int(y/self.h)
        if self.debug:
            print("get_widget_index", x, y, i)
        return i

    def set_source(self, event):
        self.source = self.get_widget_index(event)
        if self.debug:
            print("set_source", self.current)

    def move_widget_image(self, event):
        pass

    def move_to_target(self, event):
        target = self.get_widget_index(event)
        if self.debug:
            print("move_to_target", self.source, target)
        self.swap(self.source, target)

    def swap(self, i, j):
        if i == j:
            return

        source_pair = self.pair_list[i]
        target_pair = self.pair_list[j]

        self.pair_list[j] = source_pair
        source_pair[1].grid(row=j, column=self.widget_col)

        self.pair_list[i] = target_pair
        target_pair[1].grid(row=i, column=self.widget_col)

        self.update()

        if self.debug:
            for pair in self.pair_list:
                print(pair[0].get())

    def get_variables(self):
        return [pair[0] for pair in self.pair_list]



"""
    ColumnConstants.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
SECCONF_LOWER_BOUND = -2.5
BAD_CONFORMANCE_REDUCE = 1e-2       # used when it is positive
PORESIZE_BOUNDSS = (70, 100)         # deprecated and dynamically determined from Exclusion Limit

INJECTION_TIME = -1000
NUM_THEORETICAL_PLATES = 14400      # superceded by number_of_plates setting

Ti_LOWER = -5000        # -5047 with 20130303/NAC, G0665
Ti_UPPER = 0
Np_LOWER = 100
Np_UPPER = NUM_THEORETICAL_PLATES*2



# coding: utf-8
"""

    BasicUtils.py

    Copyright (c) 2016-2022, Masatsuyo Takahashi, KEK-PF

"""

from __future__ import division, print_function, unicode_literals

import sys
import os
import re
import getpass
import time
import shutil
import inspect

def mkdirs_with_retry( path, retry=3 ):
    retry_ = 0
    while( not os.path.exists( path ) ):
        try:
            os.makedirs( path )
        except:
            if retry_ < retry:
                retry_ += 1
                time.sleep( 1 )
                print( 'retry os.makedirs( %s )' % ( path ) )
            else:
                assert( False )

def clear_dirs_with_retry( dirs, retry=3  ):
    for dir_ in dirs:
        if os.path.exists( dir_ ):
            shutil.rmtree( dir_ )
        mkdirs_with_retry( dir_, retry=retry )

def rename_with_retry( old_path, new_path, retry=3 ):
    retry_ = 0
    done_ = False
    while( not done_ ):
        try:
            os.rename( old_path, new_path )
            done_ = True
        except:
            if retry_ < retry:
                retry_ += 1
                time.sleep( 1 )
                print( 'retry os.rename( %s, %s )' % ( old_path, new_path ) )
            else:
                assert( False )

def is_empty_dir( path ):
    if not os.path.exists( path ):
        return False

    for dirpath, dirnames, files in os.walk( path ):
        # print( 'files=', files )
        if files:
            return False
        else:
            return True

def num_files_in(path):
    for dirpath, dirnames, files in os.walk( path ):
        if files:
            return len(files)
        else:
            return 0

def is_almost_empty_dir( path ):
    if not os.path.exists( path ):
        return False

    for dirpath, dirnames, files in os.walk( path ):
        if len( files ) == 0:
            return True
        elif len( files ) == 1:
            path = os.path.join( dirpath, files[0] )
            return os.path.getsize( path ) == 0
        else:
            return False

def is_usedfolder(folder):
    return os.path.exists(folder) and len(os.listdir(folder)) > 0

def get_filename_extension( filename ):
    f = filename.split( '.' )
    if len( f ) > 1:
        ext = f[-1]
    else:
        ext = None
    return ext

def exe_name():
    name_ =  os.path.split( sys.argv[0] )[-1]
    for ext in ( '.exe', '.py' ):
        name_ = name_.replace( ext, '' )
    return name_

def home_dir():
    try:
        username = getpass.getuser()
    except:
        # suddenly began to happen on 20191106 on a note PC
        import win32api
        username = win32api.GetUserName()
    return os.path.expanduser( '~' + username )

"""
    URL: http://stackoverflow.com/questions/1305532/convert-python-dict-to-object

    namedtuple の場合、メンバーの更新は can't set attritube により不可のようなので
    この Struct の方が柔軟。
"""
class Struct:
    def __init__(self, **entries): 
        self.__dict__.update(entries)
    def __str__(self):
        return str(self.__dict__)

def make_indecies_text( indecies, offset=0 ):
    assert( len( indecies ) > 0 )
    previous = None
    is_continuous = True
    for i in indecies:
        if previous is not None:
            if i != previous + 1:
                is_continuous = False
        previous = i

    if is_continuous:
        if len( indecies ) == 1:
            text = str( offset+indecies[0] )
        else:
            text = '%d-%d' % ( offset+indecies[0], offset+indecies[-1] )
    else:
        text = ','.join( [ str(offset+i) for i in indecies ] )

    return text

"""
thanks to: http://stackoverflow.com/questions/9647202/ordinal-numbers-replacement
"""
def ordinal_str(numb):
    if numb < 20: #determining suffix for < 20
        if numb == 1: 
            suffix = 'st'
        elif numb == 2:
            suffix = 'nd'
        elif numb == 3:
            suffix = 'rd'
        else:
            suffix = 'th'  
    else:   #determining suffix for > 20
        tens = str(numb)
        tens = tens[-2]
        unit = str(numb)
        unit = unit[-1]
        if tens == "1":
           suffix = "th"
        else:
            if unit == "1": 
                suffix = 'st'
            elif unit == "2":
                suffix = 'nd'
            elif unit == "3":
                suffix = 'rd'
            else:
                suffix = 'th'
    return str(numb)+ suffix


def get_caller_module( level=1 ):
    # get the caller module
    frm = inspect.stack()[level]
    mod = inspect.getmodule(frm[0])
    return mod

def get_home_folder():
    folder = __file__
    for i in range( 3 ):
        folder = os.path.dirname( folder )
    return folder.replace( '\\', '/' )

def auto_numbered_file( f ):
    while os.path.exists( f ):
        f = re.sub( r'-(\d+)(\.\w+)$', lambda m: '-' + '%03d' % ( int(m.group(1)) + 1 ) + m.group(2), f )
    return f

"""
    from http://stackoverflow.com/questions/651794/whats-the-best-way-to-initialize-a-dict-of-dicts-in-python
"""

class AutoVivifiedDict(dict):
    """Implementation of perl's autovivification feature."""
    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value

def open_w_safely( path, mode, rename_cb=None ):
    assert mode == "w", "Supprted only for write mode"

    if os.path.exists( path ):
        if rename_cb is None:
            path_bak = re.sub( r"(\.\w+)$", ".bak", path )
            if os.path.exists( path_bak ):
                os.remove( path_bak )
            os.rename( path, path_bak )
        else:
            rename_cb( path )

    fh  = open( path, mode )
    return fh

def rename_existing_file(path, ext=''):
    search_pattern = "(-\\d*)?\\" + ext + "$"
    replace_pattern = "-%02d" + ext

    n = 0
    path_bak = path
    while os.path.exists( path_bak ):
        path_bak = re.sub( search_pattern, replace_pattern % n, path )
        n += 1

    if path_bak != path:
        os.rename( path, path_bak ) 

def print_exception():
    from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
    etb = ExceptionTracebacker()
    print(etb)



"""
    SSDC.SsdcAnalysis.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.interpolate import UnivariateSpline
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

class SsdcAnalysis(Dialog):
    def __init__(self, parent, info_list, num_components):
        self.parent = parent
        self.info_list = info_list
        self.num_components = num_components
        Dialog.__init__(self, parent, title="SSDC Analysis", visible=False)

    def show(self):
        self._show()
    
    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(20, 10))
        self.draw(fig, axes)
        fig.tight_layout()
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.toolbar = NavigationToolbar(self.mpl_canvas, cframe)
        self.toolbar.update()
    
    def draw(self, fig, axes):
        from importlib import reload
        import SSDC.MappingDecomposer
        reload(SSDC.MappingDecomposer)
        from SSDC.MappingDecomposer import decompose_paired_elution_curves, compute_monopore_curves

        moments_list = []
        for in_folder, lrf_src in self.info_list:
            moments_list.append(lrf_src.get_egh_moments_list())

        temp_fix_slice = slice(None, None)
        if in_folder.find("20230706/ALD_OA") >= 0:
            temp_fix_slice = slice(1, None)

        for ax, (in_folder, lrf_src), moments in zip(axes[:,0], self.info_list, moments_list):
            in_folder = get_in_folder(in_folder)
            ax.set_title("Naive Decomposition of %s" % in_folder, fontsize=16)
            model = lrf_src.model
            x = lrf_src.xr_x
            y = lrf_src.xr_y
            ax.plot(x, y, color="orange")
            for params in lrf_src.get_peaks()[temp_fix_slice]:
                cy = model(x, params)
                ax.plot(x, cy, ":")
            for M in moments[temp_fix_slice]:
                ax.axvline(x=M[0], color="blue")

        ret = decompose_paired_elution_curves(self.info_list, moments_list, self.num_components)
        if ret is None:
            return
        mnp_params, rgs, mapping, scale = ret
        print("mapping=", mapping)
        print("rgs=", rgs)

        for k, (ax, (in_folder, lrf_src), moments) in enumerate(zip(axes[:,1], self.info_list, moments_list)):
            in_folder = get_in_folder(in_folder)
            ax.set_title("Mapped Decomposition of %s" % in_folder, fontsize=16)
            model = lrf_src.model
            x = lrf_src.xr_x
            y = lrf_src.xr_y
            if k == 0:
                ax.plot(x, y, color="orange")
                cy_list, tty = compute_monopore_curves(x, mnp_params, rgs)
                peakpos = []
                for cy in cy_list:
                    ax.plot(x, cy, ":")
                    peakpos.append(x[np.argmax(cy)])
                peakpos = np.array(peakpos)
                x_ = x
                last_x = x
                times_scale = 1
                intensity_scale = 1
                volume_scale = 1
            else:
                spline = UnivariateSpline(x, y, s=0, ext=3)
                a, b  = mapping
                x_ = a*last_x + b
                y_ = spline(x_)
                ax.plot(x_, y_, color="orange")
                for cy in cy_list:
                    ax.plot(x_, scale*cy, ":")
                peakpos = a*peakpos + b
                times_scale = a
                intensity_scale = scale
                volume_scale = a*scale
            for px in peakpos:
                ax.axvline(x=px, color="blue")

            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            for k, (name, value) in enumerate([("Time Scale", times_scale), ("Intensity Scale", intensity_scale), ("Volume Scale", volume_scale)]):
                tx = xmin*0.95 + xmax*0.05
                w = 0.8 - 0.13*k
                ty = ymin*(1 - w)+ ymax*w
                ax.text(tx, ty, "%s = %.3g" % (name, value), alpha=0.3, fontsize=30)


# coding: utf-8
"""

    ファイル名：   MplPlotCanvas.py

    処理内容：

        デバッグ用の plot 

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
import numpy                            as np
from mpl_toolkits.mplot3d               import Axes3D
import matplotlib.pyplot                as plt
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter                         import Tk, Dialog

class MplPlotCanvas( Tk.Toplevel ):
    def __init__( self, title, draw_func, parent=None ):

        Tk.Toplevel.__init__( self )

        if parent is None:
            parent = Tk.Toplevel()
            parent.withdraw()
            self.created_parent = True
        else:
            self.created_parent = False

        parent.update()
        self.parent     = parent
        self.draw_func  = draw_func

        frame = Tk.Frame( self )
        frame.pack()

        self.body( frame )

    def quit( self ):
        if self.created_parent:
            self.parent.destroy()

        self.destroy()

    def body( self, body_frame ):   # overrides parent class method

        fig = plt.figure()

        self.mpl_canvas = FigureCanvasTkAgg( fig, body_frame )
        self.mpl_canvas.draw()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        # self.mpl_canvas.mpl_connect( 'button_press_event', self.button_press )
        # self.mpl_canvas.mpl_connect( 'key_press_event', self.key_press )

        self.draw_func( fig )

        # self.protocol( "WM_DELETE_WINDOW", self.quit )

    def button_press( self, *argv ):
        print( 'button_press' )
        self.mpl_canvas.draw()

    def key_press( self, *argv ):
        print( 'key_press' )
        self.mpl_canvas.draw()



"""

    GuinierPorodFit.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF

"""
import numpy            as np
from bisect             import bisect_right
from scipy              import optimize

USE_GUINIER_POROD_GENERAL = False
if USE_GUINIER_POROD_GENERAL:
    from DataModels     import GuinierPorodGeneral, GuinierPorodGeneralLmfit
else:
    from DataModels     import GuinierPorod, GuinierPorodLmfit

from lmfit              import minimize, Parameters, Parameter, report_fit
from IntensityData      import ACCEPTABLE_BASIC_QUALITY
# from molass_legacy.KekLib.NumpyUtils         import np_savetxt

DEBUG = False

class GuinierPorodFit:
    def __init__( self, intensity ):
        self.intensity  = intensity

        # if intensity.basically_ok:
        if DEBUG: print( 'basic_quality=', intensity.basic_quality )
        if intensity.basic_quality >= ACCEPTABLE_BASIC_QUALITY:
            f1, t1 = intensity.get_usable_range()
            usable_proportion = 1
        else:
            f1, t1 = intensity.get_approximate_range()
            usable_proportion = 3/4

        f1 = min( 40, f1 )
        if DEBUG: print( 'f1, t1=', [f1,t1] )

        t_ = int( intensity.Q.shape[0] * usable_proportion )

        self.Q = intensity.Q[0:t_]
        self.I = intensity.I[0:t_]
        self.X = intensity.X[0:t_]
        self.Y = intensity.Y[0:t_]
        self.W = intensity.W[0:t_]
        self.E = intensity.E[0:t_]

        self.smoother = intensity.smoother
        self.I_ = self.smoother.predict( self.Q )
        # np_savetxt( 'I_.csv', self.I_ )

        self.model = None
        self.f0 = f1
        self.t0 = t1

    def __call__( self, x, sigma=0.01 ):
        if self.model is None:
            return x * np.nan
        else:
            return self.model( x )

    def fit( self ):

        init_params = self.estimate_init_params()
        # print( 'init_params=', init_params )

        if USE_GUINIER_POROD_GENERAL:
            guinier_porod = GuinierPorodGeneralLmfit()
        else:
            guinier_porod = GuinierPorodLmfit()

        G, Rg, d = init_params

        if USE_GUINIER_POROD_GENERAL:

            min_recchi = None
            opt_result = None

            for s_init in range(3):
                d_init  = max(d, s_init+1)
                params = Parameters()
                params.add('G',     value= G,       min=0 )
                params.add('Rg',    value= Rg,      min=1.0, max=500.0 )
                params.add('d',     value= d_init,  min=1,  max=5 )
                params.add('s',     value= s_init,  min=0,  max=2 )

                result = minimize( guinier_porod, params, args=( self.Q, self.I_, self.W ) )
                G_, Rg_, d_, s_ = result.params['G'].value, result.params['Rg'].value, result.params['d'].value, result.params['s'].value

                if True:
                    print( 'G_, Rg_, d_, s_=', G_, Rg_, d_, s_ )
                    print( 'redchi=', result.redchi )

                if min_recchi is None or result.redchi < min_recchi:
                    min_recchi  = result.redchi
                    opt_result  = result

            G_, Rg_, d_, s_ = opt_result.params['G'].value, opt_result.params['Rg'].value, opt_result.params['d'].value, opt_result.params['s'].value

            if True:
                print( 'G_, Rg_, d_, s_=', G_, Rg_, d_, s_ )
                print( 'redchi=', opt_result.redchi )

            self.result = opt_result
            self.model  = GuinierPorodGeneral( G_, Rg_, d_, s_ )

        else:

            params = Parameters()
            params.add('G',     value= G,   min=0 )
            params.add('Rg',    value= Rg,  min=1.0, max=500.0 )
            params.add('d',     value= d,   min=1,  max=5 )

            result = minimize( guinier_porod, params, args=( self.Q, self.I_, self.W ) )
            G_, Rg_, d_ = result.params['G'].value, result.params['Rg'].value, result.params['d'].value

            self.result = result
            self.model  = GuinierPorod( G_, Rg_, d_ )

        self.I0     = G_
        self.Rg     = Rg_
        self.d      = d_
        self.degree = d_

        # TODO: refactoring
        self.eps    = result.residual
        self.q1     = 1/Rg_ * np.sqrt( 3*d_/2 )
        self.t1     = bisect_right( self.Q, self.q1 )

        self.cover_ratio = 1.0      # TODO: revise it
        cover_  = self.cover_ratio * 100
        self.remarks = 'Guinier-Porod Fit Info: Rg=%.3g, q1=%.3g, q1*Rg=%.3g, d=%.3g, interval=[%d, %d], cover=%d%%' % (
                        self.Rg, self.q1, self.Rg*self.q1, self.degree, self.f0, self.t0, cover_ )


    def equations( self, p ):
        G, Rg, d = p

        if not ( d>=1 and d <= 5 and d>=1 and d <= 5 ):
            return ( np.inf, ) * 6

        if G < 0:
            return ( np.inf, ) * 6

        if Rg <= 0:
            return ( np.inf, ) * 6

        qq  = self.Q_array[0]**2
        f0  = G * np.exp( - qq * Rg**2 / 3 ) - self.I_array[0]

        qq  = self.Q_array[1]**2
        f1  = G * np.exp( - qq * Rg**2 / 3 ) - self.I_array[1]

        q   = np.sqrt( self.Q_array[2] )
        D   = G * np.exp( -d/2 ) * np.power( 3*d/2, d/2 ) / np.power( Rg, d )
        f2  = D / q**d + D / q**d - self.I_array[2]

        return ( f0, f1, f2 )
        # determine G, Rg, d so that this returns ( 0, 0, 0 ) 

    def estimate_init_params( self ):

        # print( 't0=', self.t0 )

        q0  = self.Q[0]
        q1  = self.Q[40]        # TODO: determine a safe value for 20
        t_  = min( 80, self.t0 )
        q2  = self.Q[ t_ ]

        Q_ = np.array( [ q0,  q1,  q2 ] )
        self.Q_array = Q_
        self.I_array = self.smoother.predict( Q_ )

        def f(p):
            return np.abs( np.sum( np.array( self.equations( p ) )**2 ) - 0 )

        G_init = self.I_array[0] / np.exp( -q0**2 * 30.0**2/3 )
        p_init = optimize.fmin( f, ( G_init, 30.0, 2.5 ), disp=False )

        """
        # it suffices to optimize.fmin and no need to optimize.fsolve
        G_, Rg_, d_ = optimize.fsolve( self.equations, p_init )
        print( 'estimate_init_params: ', G_, Rg_, d_ )
        """

        return p_init



"""
    Solvers.UltraNest.SamplerCallback.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import sys
from ultranest.viz import get_default_viz_callback
from molass_legacy._MOLASS.Version import is_developing_version

def get_viz_callback():
    from importlib import reload
    import Solvers.UltraNest.CustomLivePointsWidget
    reload(Solvers.UltraNest.CustomLivePointsWidget)
    from Solvers.UltraNest.CustomLivePointsWidget import CustomLivePointsWidget

    default_callback = get_default_viz_callback()
    if is_developing_version():
        return CustomLivePointsWidget(default_callback=default_callback)
    else:
        return default_callback

class StderrWinmode:
    def __init__(self):
        pass

    def isatty(self):
        return False
    
    def write(self, *args):
        pass

    def flush(self):
        pass

class StdoutWinmode:
    def __init__(self):
        pass

    def isatty(self):
        return False
    
    def write(self, *args):
        pass

    def flush(self):
        pass

class SamplerCallback:
    def __init__(self, solver, sampler):
        self.default_callback = get_viz_callback()
        self.callback = solver.callback
        self.sampler = sampler
        self.counter = 0
        if sys.stderr is None:
            sys.stderr = StderrWinmode()    # for sys.stderr.isatty() in ultranest.viz.py on Windows win-app
        if sys.stdout is None:
            sys.stdout = StdoutWinmode()    # for sys.stdout.write(...) in ultranest.integrator.py on Windows win-app

    def __call__(self, points, info, region, transformLayer, region_fresh=False):
        self.default_callback(points, info, region, transformLayer, region_fresh=region_fresh)

        print("SamplerCallback.__call__: info.keys()=", info.keys())
        stepsampler_info = info['stepsampler_info']
        if len(stepsampler_info) > 0:
            print("SamplerCallback.__call__: stepsampler_info.keys()=", stepsampler_info.keys())
 
        if self.sampler.results is not None:
            print("SamplerCallback.__call__: results.keys()=", self.sampler.results.keys())
        
        self.counter += 1
        print("SamplerCallback.__call__: counter=", self.counter)
        print("SamplerCallback.__call__: points.keys()=", points.keys())
        print("SamplerCallback.__call__: points['u'].shape=", points['u'].shape)
        print("SamplerCallback.__call__: points['p'].shape=", points['p'].shape)
        print("SamplerCallback.__call__: points['logl'].shape=", points['logl'].shape)
        m = np.argmax(points['logl'])
        print("SamplerCallback.__call__: m=", m)
        print("SamplerCallback.__call__: points['u'][m]=", points['u'][m])
        print("SamplerCallback.__call__: points['p'][m]=", points['p'][m])
        self.callback(points['p'][m], None, False)


"""

    DatgnomResultBook.py

    Copyright (c) 2018-2023, SAXS Team, KEK-PF

"""
import os
import numpy                as np
import csv
import logging
from datetime               import datetime
# from openpyxl               import Workbook
from openpyxl.styles        import Alignment
from openpyxl.utils         import get_column_letter
from openpyxl.chart         import ScatterChart, Reference, Series
from openpyxl.chart.data_source     import StrRef
from openpyxl.chart.layout  import Layout, ManualLayout
from OpenPyXlUtil           import save_allowing_user_reply
from molass_legacy._MOLASS.Version                import get_version_string
from molass_legacy._MOLASS.SerialSettings         import get_setting
from LinearityScore         import linearity_score100, stderror_score100, FACTOR_WEIGHT
from molass_legacy.Reports.DefaultFont import set_default_font
set_default_font()

CHART_WIDTH     = 25
CHART_HEIGHT    = 15
CHART_1_POS     = 'K3'
CHART_2_POS     = 'K35'

def make_datgnom_result_book(
            excel_is_available, wb, ws,
            mjn, range_type, datgnom_out_file, book_file, parent ):

    book = DatgnomResultBook( wb, ws, mjn, range_type, datgnom_out_file, parent )

    if excel_is_available:
        book.save( book_file )
        book.add_format_setting( book_file )
    else:
        _, file = os.path.split(book_file)
        parent.logger.warning( 'excel book formatting has been skipped for %s.', file )

class DatgnomResultBook:
    def __init__( self, wb, ws, mjn, range_type, datgnom_out_file, parent ):
        self.logger = logging.getLogger( __name__ )

        m, j, num_peaks_to_exec, peak_num_ranges = mjn
        self.m  = m
        self.ud = j
        self.wb = wb
        self.ws = ws
        self.parent = parent

        paren = '' if num_peaks_to_exec == 1 else '(%d)' % ( m + 1 )
        if peak_num_ranges == 1 and range_type < 5:
            side = 'Both'
        else:
            side = 'Asc' if self.ud == 0 else 'Desc'
        ws.title = '%s-side Datgnom Result%s' % ( side, paren )

        ws.append( [ None, 'Experimental Data and Fit', None, None, None, None, 'Real Space Data' ]  )
        ws.append( [ 'S', 'J EXP', 'ERROR', 'J REG', 'I REG', None, 'R', 'P(R)', 'ERROR' ] )

        exper_fit, real_space = self.read_data( datgnom_out_file )

        self.num_rows_exper_fit     = len(exper_fit)
        self.num_rows_real_space    = len(real_space)
        num_rows = max( self.num_rows_exper_fit , self.num_rows_real_space )

        for i in range( num_rows ):
            if i < len(exper_fit):
                row = exper_fit[i]
            else:
                row = [None] * 5

            row += [None]

            if i < len(real_space):
                row += real_space[i]
            else:
                row += [None] * 3

            ws.append( row )

        self.create_scattering_curve_chart()
        self.create_pr_function_chart()

    def save( self, xlsx_file ):
        save_allowing_user_reply( self.wb, xlsx_file )

    def read_data(self, datgnom_out_file):
        from molass_legacy.ATSAS.DatGnom import datgnom_read_data
        return datgnom_read_data(datgnom_out_file)

    def create_scattering_curve_chart( self ):
        c_ = ScatterChart()
        c_.title = "Datgnom generated scattering intensity"
        c_.y_axis.title = 'log( I )'
        c_.x_axis.title = 'Q(Å⁻¹)'

        min_row = 2
        max_row = 2 + self.num_rows_exper_fit
        xvalues = Reference(self.ws, min_col=1,
                                min_row=min_row + 1,    # must not include column title
                                max_row=max_row )

        for col in [ 2, 5 ]:
            values = Reference(self.ws, min_col=col, min_row=min_row, max_row=max_row)
            series = Series(values, xvalues, title_from_data=True)
            c_.series.append(series)

        c_.y_axis.scaling.logBase = 10

        c_.width    = CHART_WIDTH
        c_.height   = CHART_HEIGHT

        adjust = 0
        c_.layout = Layout(
            ManualLayout(
            x=0.07+adjust, y=0.11,
            h=0.8,  w=0.86+adjust,
            xMode="edge",
            yMode="edge",
            )
        )

        self.ws.add_chart( c_, CHART_1_POS )

    def create_pr_function_chart( self ):
        c_ = ScatterChart()
        c_.title = "Datgnom generated P(r)"
        c_.y_axis.title = 'p(r)'
        c_.x_axis.title = 'r'

        min_row = 2
        max_row = 2 + self.num_rows_real_space
        xvalues = Reference(self.ws, min_col=7,
                                min_row=min_row + 1,    # must not include column title
                                max_row=max_row )

        for col in [ 8 ]:
            values = Reference(self.ws, min_col=col, min_row=min_row, max_row=max_row)
            series = Series(values, xvalues, title_from_data=True)
            c_.series.append(series)

        c_.width    = CHART_WIDTH
        c_.height   = CHART_HEIGHT

        adjust = 0
        c_.layout = Layout(
            ManualLayout(
            x=0.07+adjust, y=0.11,
            h=0.8,  w=0.86+adjust,
            xMode="edge",
            yMode="edge",
            )
        )

        self.ws.add_chart( c_, CHART_2_POS )

    def add_format_setting( self, book_file ):
        from .DatgnomExcelFormatter import DatgnomResultArgs
        args = DatgnomResultArgs(self.ws.title, book_file)

        if self.parent.more_multicore:
            self.add_format_setting_more_multicore(args)
        else:
            self.add_format_setting_less_multicore(args)

    def add_format_setting_more_multicore( self, args ):
        self.parent.teller.tell('range_datgnom_book', args=args)

    def add_format_setting_less_multicore( self, args ):
        from .DatgnomExcelFormatter import add_datgnom_format_setting

        add_datgnom_format_setting(self.parent.excel_client, args, self.logger)



"""
    RangeEditorFrame.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
import os
import copy
import numpy                as np
import matplotlib.pyplot    as plt
from matplotlib.gridspec    import GridSpec
import matplotlib.patches   as mpl_patches      # 'as patches' does not work properly
from matplotlib import colors
import matplotlib
import seaborn as sns
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk
from molass_legacy.KekLib.TkUtils                import split_geometry
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar, get_color, get_hex_color
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting
from DevSettings            import get_dev_setting
from .SuperEditorFrame import SuperEditorFrame
from molass_legacy.DataStructure.AnalysisRangeInfo      import AnalysisRangeInfo, convert_to_paired_ranges
from RangeInfo              import RangeEditorInfo, shift_editor_ranges
from PairedRangeLogger      import log_paired_ranges
from .RangeSpecFrame import RangeSpecFrame, RANGE_FRAME_HEIGHT

MIN_NUM_PANEL_SPACES    = 3
CANVAS_HIGHT    = 6
RANGE_EDITOR_LOG_HEADER  = "--- range editor log begin ---"
RANGE_EDITOR_LOG_TRAILER = "--- range editor log end ---"

def format_coord(x, y):
    return 'x=%.4g    y=%.4g' % (x, y)

class RangeEditorFrame(SuperEditorFrame):
    def __init__( self, parent, dialog, sd, mapper, corbase_info ):
        self.dialog = dialog
        self.logger = dialog.logger
        Tk.Frame.__init__(self, parent)

        self.mplt_ge_2_2 = matplotlib.__version__ >= '2.2'

        self.mapper = mapper
        self.xr_j0 = sd.xr_j0
        self.corbase_info = corbase_info

        self.elution_fig_type = 0   # 0 : Xray, 1:UV
        self.popup_menu = None
        self.ex_solver = None

        e_curve = mapper.x_curve
        self.xr_x = self.xr_j0 + e_curve.x
        self.xr_y = mapper.x_curve_y_adjusted
        self.uv_y = mapper.mapped_vector
        self.peaks = e_curve.get_emg_peaks()
        self.get_editor_ranges(e_curve, self.peaks)
        self.num_peaks = len(self.peaks)

        self.recompute_decomposition(redraw=False)

        self.params_controllable = False

        self.toggle = False
        title = "Range Specification for %s" % get_setting('in_folder')
        SuperEditorFrame.__init__(self, parent, title)

    def get_editor_ranges(self, curve, peaks):
        range_info = get_setting('range_editor_info')
        self.is_memorized = range_info is not None      # just temporary
        default_ranges = curve.get_default_editor_ranges()

        if range_info is None:
            ranges = default_ranges
            ignorable_flags = [0] * len(ranges)
        else:
            range_info.update(default_ranges)
            ranges = range_info.get_ranges()
            ignorable_flags = range_info.get_ignorable_flags()

        assert len(ranges) == len(peaks)

        self.editor_ranges = shift_editor_ranges(self.xr_j0, ranges)
        self.ignorable_flags = ignorable_flags

    def debug_plot(self):
        from molass_legacy.DataStructure.AnalysisRangeInfo import AnalysisRangeInfo
        from DecompEditorDebug import debug_plot_impl
        ranges = AnalysisRangeInfo(self.make_range_info()).get_ranges()
        debug_plot_impl(self.x, ranges, np.arange(253, 302))

    def get_figsize(self):
        fig_height = min( CANVAS_HIGHT, self.num_peaks*2 )
        return (3, fig_height)

    def update_error_label(self):
        # required to avoid accessing to self.fit_error
        pass

    def draw1(self):
        ax  = self.ax1
        ax.cla()

        x   = self.xr_x
        xr_y = self.xr_y
        uv_y = self.uv_y

        ax.plot(x, xr_y, color='orange', label='Adjusted Xray Elution')
        ax.plot(x, uv_y, ':', color='blue', label='Mapped UV Elution')

        ax.legend()
        self.fig1_range_parts = []
        self.add_range_patchs(ax, self.fig1_range_parts)

        # draw the difference
        ax1r = self.ax1r
        ax1r.cla()

        diff_y = uv_y - xr_y
        ax1r.bar( x, diff_y, color='purple', label='Difference (UV - Xray)', alpha=0.5 )
        ax1r.plot( x[[0, -1]], [0, 0], color='pink' )

        ymin, ymax = ax.get_ylim()
        yminr, ymaxr = ax1r.get_ylim()
        ymidr = (yminr + ymaxr)/2
        half_height_r = max((ymax - ymin)*self.height_ratio, ymaxr - yminr)/2
        ax1r.set_ylim(ymidr - half_height_r, ymidr + half_height_r)

        ax1r.legend(bbox_to_anchor=(1, 1), loc='lower right')
        self.fig1_range_parts_resid = []
        self.add_range_patchs(ax1r, self.fig1_range_parts_resid)

    def draw2(self, peak_no=None):

        if self.axes is not None:
            self.fig2.clf()
            figsize = self.get_figsize()
            self.fig2.set_size_inches(*figsize)

        x   = self.xr_x
        xr_y = self.xr_y
        uv_y = self.uv_y

        num_peaks = self.num_peaks
        gs = GridSpec( num_peaks, 1 )
        self.axes = [ self.fig2.add_subplot( gs[i] ) for i in range(num_peaks) ]

        self.fig2_range_parts_list = []
        for k, ax in enumerate(self.axes):
            if peak_no is not None:
                # do the following only for k == i
                if k != peak_no - 1:
                    # continue
                    pass

            ax.cla()
            # ax.set_axis_off()
            ax.axes.get_xaxis().set_ticks([])
            ax.axes.get_yaxis().set_ticks([])
            ax.plot(x, xr_y, color='orange')
            ax.plot(x, uv_y, ':', color='blue')

            range_parts = []

            if  ( len(self.specpanel_list) == 0 and not self.ignorable_flags[k]
                or len(self.specpanel_list) > 0 and self.specpanel_list[k].ignore.get() == 0 ):
                self.add_range_patchs(ax, range_parts, k=k)

            self.fig2_range_parts_list.append(range_parts)

    def add_range_patchs(self, ax, range_parts, k=None):
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)

        range_parts.append([ymin, ymax])

        if k is None:
            editor_ranges_ = self.editor_ranges
        else:
            editor_ranges_ = [self.editor_ranges[k]]

        for j, range_list in enumerate(editor_ranges_):
            if k is None and self.ignorable_flags[j]:
                range_parts.append([j])
                continue

            for f, t in range_list:
                p = mpl_patches.Rectangle(
                        (f, ymin),  # (x,y)
                        t - f,   # width
                        ymax - ymin,    # height
                        facecolor   = 'cyan',
                        alpha       = 0.2,
                    )
                ax.add_patch( p )
                lines = []
                for x in [f, t]:
                    line, = ax.plot( [x, x], [ymin, ymax], ':', color='gray', alpha=0.5 )
                    lines.append(line)
                range_parts.append([j, p, lines])

    def refresh_figs(self):
        self.draw1()
        self.draw2()
        self.canvas_draw1()
        self.canvas_draw2()

    def update1(self):
        print('update1 start')
        try:
            self.update1_impl(self.fig1_range_parts)
            self.update1_impl(self.fig1_range_parts_resid)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print(etb)
        print('update1 end')

    def update1_impl(self, range_parts):
        print('update1_impl start')
        ymin, ymax = range_parts[0]
        for k, parts in enumerate(range_parts[1:]):
            if parts is None:
                continue

            j, patch, lines = parts
            range_list = self.editor_ranges[j]
            print([k, j], 'range_list=', range_list)
            for f, t in range_list:
                x, y = patch.xy
                patch.set_xy((f, y))
                patch.set_width(t - f)
                for n, x_ in enumerate([f, t]):
                    line = lines[n]
                    line.set_data([x_, x_], [ymin, ymax])

    def update2(self, peak_no):
        print('update2: peak_no=', peak_no)
        pk = peakno - 1
        range_parts = self.fig2_range_parts_list[pk]

        ymin, ymax = range_parts[0]
        for k, parts in enumerate(range_parts[1:]):
            if parts is None:
                continue

            j, patch, lines = parts
            range_list = self.editor_ranges[j]
            for f, t in range_list:
                x, y = patch.xy
                patch.set_xy((f, y))
                patch.set_width(t - f)

    def update_figs(self, peak_no):
        self.update1()
        self.update2(peak_no)
        self.canvas_draw1()
        self.canvas_draw2()

    def add_widgets(self):
        num_peaks = self.num_peaks

        canvas_width = int(self.mpl_canvas2_widget.cget('width')) + self.pcframe_padx*2
        self.col_widths  = [canvas_width, 60, 80, 60, 160, 60, 60, 160]
        label_texts = [ 'Figure',
                        'Peak No',
                        'Number of\nRanges',
                        'Range Id',
                        'Range(s)\nFrom       To',
                        'Ignore',
                        ]

        self.col_title_frames = []
        for k, text in enumerate(label_texts):
            frame = Tk.Frame(self.ptframe, width=self.col_widths[k], height=40)
            frame.grid(row=0, column=k, sticky=Tk.S)
            self.col_title_frames.append(frame)
            # https://stackoverflow.com/questions/16363292/label-width-in-tkinter
            frame.pack_propagate(0)
            label = Tk.Label(frame, text=text)
            label.pack(side=Tk.BOTTOM)

        for k in range(max(MIN_NUM_PANEL_SPACES, num_peaks)):
            self.ppframe.grid_rowconfigure(k, weight=1)

        self.refresh_spec_panels()

        for k in range( max(0, MIN_NUM_PANEL_SPACES - num_peaks) ):
            space = Tk.Frame(self.ppframe, height=RANGE_FRAME_HEIGHT)
            space.grid(row=num_peaks+k, column=0)

    def make_range_info(self, concentration_datatype=3):
        # concentration_datatype is not used here.
        # it is included in the above arguments to keep the inferface consistent
        # as seen from PreviewButtonFrame
        paired_ranges = []
        for k, panel in enumerate(self.specpanel_list):
            if self.ignorable_flags[k]:
                continue
            paired_ranges.append(panel.get_paired_range())
        return paired_ranges

    def make_restorable_ranges(self):
        paired_ranges = []
        for panel in self.specpanel_list:
            paired_ranges.append(panel.get_paired_range())
        return paired_ranges

    def apply( self ):  # overrides parent class method
        set_setting( 'use_elution_models', 0 )

        analysis_range_info = AnalysisRangeInfo( self.make_range_info(), editor='RangeEditor' )
        print('analysis_range_info=', analysis_range_info)
        set_setting( 'analysis_range_info', analysis_range_info )
        set_setting( 'range_type', 4 )
        paired_ranges = self.make_restorable_ranges()
        set_setting( 'range_editor_info', RangeEditorInfo(paired_ranges, self.ignorable_flags))

        ret = convert_to_paired_ranges(analysis_range_info.get_ranges())
        paired_ranges = ret[0]
        self.log_info_for_reset()
        log_paired_ranges(self.logger, paired_ranges)

    def set_data_label(self):
        pass

    def set_toggle_text(self):
        pass

    def toggle_show( self ):
        self.elution_fig_type = 1 - self.elution_fig_type
        self.refresh_frame()

    def refresh_frame(self, elution_fig_type=None):
        if elution_fig_type is not None:
            self.elution_fig_type = elution_fig_type
        self.set_data_label()
        self.set_toggle_text()
        self.refresh_figs()

    def reset_to_defaults( self ):
        pass

    def toggle_params_constraints( self ):
        btn_text = self.param_btn.cget('text')
        is_show_button = btn_text.find("Show") >= 0
        if is_show_button:
            k = len(self.col_title_frames) - 1
            self.col_title_frames[-2].grid(row=0, column=k-1, sticky=Tk.S)
            self.col_title_frames[-1].grid(row=0, column=k, sticky=Tk.S)
        else:
            self.col_title_frames[-2].grid_forget()
            self.col_title_frames[-1].grid_forget()

        for specpanel in self.specpanel_list:
            if is_show_button:
                specpanel.constraints_restore()
            else:
                specpanel.constraints_forget()

        if is_show_button:
            text_ = btn_text.replace("Show", "Hide")
            self.recompute_btn.grid(row=0, column=3, padx=5)
        else:
            text_ = btn_text.replace("Hide", "Show")
            self.recompute_btn.grid_forget()

        self.param_btn.config(text=text_)

        self.update()   # necessary to make effective the following xview_moveto

        if is_show_button:
            if self.dialog.scrollable:
                self.dialog.scrolled_frame.canvas.xview_moveto(1)

    def recompute_decomposition( self, redraw=True ):

        self.specpanel_list = []

        if redraw:
            self.refresh_figs()
            self.refresh_spec_panels(restore=True)
            self.panel_frame.update()

    def refresh_spec_panels(self, restore=False):
        if len(self.specpanel_list) > 0:
            for specpanel in self.specpanel_list:
                specpanel.destroy()
            self.specpanel_list = []

        active_peak_no = 0
        row_no_base = 0
        for k in range(self.num_peaks):
            # print('editor_ranges[%d]=' % k, self.editor_ranges[k])
            if not self.ignorable_flags[k] and len(self.editor_ranges[k]) > 0:
                active_peak_no += 1

            range_list = self.editor_ranges[k]
            panel = RangeSpecFrame(self.ppframe,
                            editor=self,
                            j_min=self.xr_j0,
                            j_max=self.xr_j0 + len(self.xr_x) - 1,
                            peak_no=k+1,
                            active_peak_no=None if self.ignorable_flags[k] else active_peak_no,
                            row_no_base=row_no_base,
                            col_widths=self.col_widths,
                            range_list=range_list,
                            peak=self.peaks[k],
                            )
            if not self.ignorable_flags[k]:
                row_no_base += len(range_list)
            if restore:
                panel.constraints_restore()
            panel.grid(row=k, column=0, sticky=Tk.N+Tk.S)
            self.specpanel_list.append(panel)

    def log_info_for_reset(self, modification="last applied"):
        self.logger.info(RANGE_EDITOR_LOG_HEADER)

        self.logger.info(modification + " model_name=None")
        self.logger.info("used ranges:")
        for k, panel in enumerate(self.specpanel_list):
            if self.ignorable_flags[k] or len(self.editor_ranges[k]) == 0:
                continue

            info = panel.get_info_for_reset()
            self.logger.info(str([k+1]) + ' ' + str(info[1]) )

        self.logger.info(RANGE_EDITOR_LOG_TRAILER)

    def on_mpl_button_press(self, event):
        pass

    def update_range(self, pno, ad, f, t):
        print('update_range:', pno, ad, f, t)
        range_ = self.editor_ranges[pno][ad]
        range_[0] = f
        range_[1] = t
        panel = self.specpanel_list[pno]
        panel.update_range(ad, f, t)
        # self.update_figs(pno)
        self.refresh_figs()

    def get_extrapolation_solver(self):
        if self.ex_solver is None:
            """
            self.ex_solver should be set to None
            whenever popts are changed.
            """
            from ExtrapolationSolver import ExtrapolationSolver
            pdata, popts = self.dialog.preview_frame.get_preview_data()
            self.ex_solver = ExtrapolationSolver(pdata, popts)
        return self.ex_solver



# basin hopping global optimization for the ackley multimodal objective function
import logging
import numpy as np
from numpy.random import rand
from numpy import exp
from numpy import sqrt
from numpy import cos
from numpy import e
from numpy import pi
from numpy import meshgrid
from numpy import arange
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon
from matplotlib import animation
from ultranest import ReactiveNestedSampler 

# objective function
def objective(v):
    x, y = v
    return -20.0 * exp(-0.2 * sqrt(0.5 * (x**2 + y**2))) - exp(0.5 * (cos(2 * pi * x) + cos(2 * pi * y))) + e + 20

DIM = 2
# define range for input
r_min, r_max = -5.0, 5.0

lower = np.ones(2) * r_min
upper = np.ones(2) * r_max
def my_prior_transform(cube):
    # transform location parameter: uniform prior
    params = cube * (upper - lower) + lower
    return params

nfev = 0
def my_likelihood(params):
    global nfev
    # print("objective_func_wrapper: par=", par)
    fv = objective(params)
    return -fv

vis_info_list = []
init_points = [(-4, -4), (-4, -4), (4, -4), (4, -4), (4, 4), (4, 4), (-4, 4), (-4, -4)]
# init_points = [(-4, -4)]

def solve(k):
    pt = init_points[k]
    fv = objective(pt)
    vis_info_list.append((pt, fv))

    ncall = 0
    def viz_callback(points, info, region, transformLayer, region_fresh=False):
        nonlocal ncall
        ncall += 1
        if ncall < 20:
            logl = points['logl']
            m = np.argmax(logl)
            x = points['p'][m]
            fv = objective(x)
            vis_info_list.append((x, fv))

    param_names = ["p%d" % i for i in range(DIM)]
    sampler = ReactiveNestedSampler(param_names, my_likelihood, my_prior_transform)
    sampler.logger.setLevel(logging.INFO)       # to suppress debug log    

    result = sampler.run(min_num_live_points=64, max_ncalls=1000, viz_callback=viz_callback)
    opt_params = result['maximum_likelihood']['point']

    # perform the basin hopping search
    return dict(x=opt_params, nfev=nfev, ncall=ncall, message="ok")

history_pos = 0
for k in range(len(init_points)):
    result = solve(k)

    # summarize the result
    print('Status : %s' % result['message'])
    print('Total Evaluations: %d' % result['nfev'])
    print('Callback calls: %d' % result['ncall'])

    # evaluate solution
    solution = result['x']
    evaluation = objective(solution)
    print('Solution: f(%s) = %.5f' % (solution, evaluation))

# sample input range uniformly at 0.1 increments
xaxis = arange(r_min, r_max, 0.1)
yaxis = arange(r_min, r_max, 0.1)
# create a mesh from the axis
xx, yy = meshgrid(xaxis, yaxis)
# compute targets
results = objective((xx, yy)) 
# create a surface plot with the jet color scheme
fig = plt.figure(figsize=(16,8))
ax = fig.add_subplot(121, projection='3d')
ax2 = fig.add_subplot(122)

fig.suptitle("Demonstration of Nested Sampling Algorithm using Ackley function", fontsize=20)
ax.set_title("Surface Plot", fontsize=16)
ax2.set_title("Contour Plot", fontsize=16)
fig.tight_layout()

surface = ax.plot_surface(xx, yy, results, cmap='jet', alpha=0.3)
contuor = ax2.contourf(xx, yy, results, cmap='jet', alpha=0.3)
# show the plot

minima_array = np.array([np.concatenate([x, [v]]) for x, v in vis_info_list])
x, y, z = minima_array[0]
points, = ax.plot([x], [y], [z], 'o', color='cyan')
trace3, = ax.plot(*minima_array[0:1,:].T, color='red')
point, = ax2.plot(x, y, 'o', color='cyan')
trace2, = ax2.plot(*minima_array[0:1,0:2].T, color='red', label="local minima trace")
ax2.legend()

def reset():
    x, y, z = minima_array[0]
    points.set_data(np.array([x]), np.array([y]))
    points.set_3d_properties(np.array([z]), 'z')
    trace3.set_data(*minima_array[0:1,0:2].T)
    trace3.set_3d_properties(minima_array[0:1,2], 'z')
    point.set_data([x], [y])
    trace2.set_data(*minima_array[0:1,0:2].T)
    return points, trace3, point, trace2

def animate(i):
    global step_arrow

    x, y, z = minima_array[i]
    points.set_data(np.array([x]), np.array([y]))
    points.set_3d_properties(np.array([z]), 'z')
    point.set_data([x], [y])
    color = 'cyan' if i % 20 == 0 else 'red'
    points.set_color(color)
    point.set_color(color)
    j, r = divmod(i, 20)
    start = j*20
    stop = start + r + 1
    slice_ = slice(start, stop)
    trace3.set_data(*minima_array[slice_,0:2].T)
    trace3.set_3d_properties(minima_array[slice_,2], 'z')
    trace2.set_data(*minima_array[slice_,0:2].T)
    return points, trace3, point, trace2

anim = animation.FuncAnimation(fig, animate, frames=len(vis_info_list), blit=True, init_func=reset, interval=500)
import sys
if len(sys.argv) > 1 and sys.argv[1] == "save":
    anim.save("ns-anim.gif")

plt.show()



"""
    Solvers.UltraNest.GridspecLayoutMpl.py

    adapted from ipywidgets.GridspecLayout

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

class GridspecLayoutMpl:
    def __init__(self, nrows, ncols):
        from Solvers.UltraNest.ProgressCanvas import ProressCanvasClient
        self.nrows = nrows
        self.ncols = ncols
        self.data = np.zeros((nrows,ncols))
        self.canvas = ProressCanvasClient(nrows,ncols)
        self.canvas.update(self.data)

    def close(self):
        self.canvas.quit()

    def _get_indices_from_slice(self, row, column):
        "convert a two-dimensional slice to a list of rows and column indices"
        # borrowd from ipywidgets.GridspecLayout

        if isinstance(row, slice):
            start, stop, stride = row.indices(self.nrows)
            rows = range(start, stop, stride)
        else:
            rows = [row]

        if isinstance(column, slice):
            start, stop, stride = column.indices(self.ncols)
            columns = range(start, stop, stride)
        else:
            columns = [column]

        return rows, columns

    def __setitem__(self, key, value):
        row, column = key

        rows, columns = self._get_indices_from_slice(row, column)
        self.data[rows,columns] = value
        self.canvas.update(self.data)

    def __getitem__(self, key):
        rows, columns = self._get_indices_from_slice(*key)
        return self.data[rows,columns]

if __name__ == '__main__':
    import os
    import sys
    sys.path.append(os.path.dirname(os.path.dirname((os.path.dirname(__file__)))))
    from time import sleep
    grid = GridspecLayoutMpl(6,10)
    grid[3,:] = 1
    grid[:,5] = 1
    grid[4,4] = 1
    print(grid.data)
    for i in range(10):
        sleep(1)
        grid[5,i] = 0.5
    sleep(3)
    grid.close()


"""
    SecTheory.ExclusionCurveSpike.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np
from matplotlib.gridspec import GridSpec

def compute_tr_impl(rgv, proportions, poresizes, Ps, x0, m):
    proportions_ = proportions/np.sum(proportions)

    PKsec = 1
    for p, rp, P in zip(proportions_, poresizes, Ps):
        rho = rgv/rp
        rho[rho > 1] = 1
        PKsec *= (P*(1 - rho)**m)**p

    return x0 + PKsec

class ExclusionCurvepy:
    def __init__(self, poresizes, proportions, P, x0, m):
        self.poresizes = poresizes
        self.proportions = proportions
        self.x0 = x0
        self.P = P
        self.m = m

    def compute_deviation(self, rg_points):
        pass

    def compute_tr(self, rgv):
        pass

    def fit(self, rg_points):
        pass

def demo():
    import molass_legacy.KekLib.DebugPlot as plt

    rp = 150

    rgv = np.flip(np.linspace(0, rp, 100))

    x0 = 50
    P = 200
    m = 3

    proportions1 = np.ones(1)
    poresizes1 = np.array([rp])
    Ps1 =  np.array([P])

    proportions2 = np.ones(2)/2
    poresizes2 = np.array([rp]*2)
    Ps2 =  np.array([P]*2)

    proportions3 = np.ones(3)/3
    poresizes3 = np.array([rp]*3)
    Ps3 =  np.array([P]*3)

    proportions4 = np.ones(2)/2
    poresizes4 = np.array([100, 100])
    Ps4 =  np.array([P, P])

    proportions5 = np.ones(2)/2
    poresizes5 = np.array([150, 50])
    Ps5 =  np.array([P/2, P*2])

    proportions6 = np.ones(2)/2
    poresizes6 = np.array([150, 50])
    Ps6 =  np.array([P, P])

    tr1 = compute_tr_impl(rgv, proportions1, poresizes1, Ps1, x0, m)
    tr2 = compute_tr_impl(rgv, proportions2, poresizes2, Ps2, x0, m)
    tr3 = compute_tr_impl(rgv, proportions3, poresizes3, Ps3, x0, m)
    tr4 = compute_tr_impl(rgv, proportions4, poresizes4, Ps4, x0, m)
    tr5 = compute_tr_impl(rgv, proportions5, poresizes5, Ps5, x0, m)
    tr6 = compute_tr_impl(rgv, proportions6, poresizes6, Ps6, x0, m)

    with plt.Dp():
        fig, ax = plt.subplots()

        ax.plot(tr1, rgv, label="x0:%g, poresizes: %s" % (x0, str(poresizes1)))
        ax.plot(tr2, rgv, label="x0:%g, poresizes: %s" % (x0, str(poresizes2)))
        ax.plot(tr3, rgv, label="x0:%g, poresizes: %s" % (x0, str(poresizes3)))
        ax.plot(tr4, rgv, label="x0:%g, poresizes: %s" % (x0, str(poresizes4)))
        ax.plot(tr5, rgv, label="x0:%g, poresizes: %s" % (x0, str(poresizes5)))
        ax.plot(tr6, rgv, label="x0:%g, poresizes: %s" % (x0, str(poresizes6)))

        ax.legend()
        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()
    import molass_legacy.KekLib

    demo()
